# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import TaskPackageDropbox

##__________________________________________________________________||
@pytest.fixture()
def workingarea():
    return mock.MagicMock()

@pytest.fixture()
def dispatcher():
    return mock.MagicMock()

@pytest.fixture()
def obj(workingarea, dispatcher):
    ret = TaskPackageDropbox(workingArea=workingarea, dispatcher=dispatcher, sleep=0.01)
    ret.open()
    yield ret
    ret.close()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_open_terminate_close(workingarea, dispatcher):

    obj = TaskPackageDropbox(workingArea=workingarea, dispatcher=dispatcher, sleep=0.01)

    assert 0 == workingarea.open.call_count
    assert 0 == workingarea.close.call_count
    assert 0 == dispatcher.terminate.call_count

    obj.open()
    assert 1 == workingarea.open.call_count
    assert 0 == workingarea.close.call_count
    assert 0 == dispatcher.terminate.call_count

    obj.terminate()
    assert 1 == workingarea.open.call_count
    assert 0 == workingarea.close.call_count
    assert 1 == dispatcher.terminate.call_count

    obj.close()
    assert 1 == workingarea.open.call_count
    assert 1 == workingarea.close.call_count
    assert 1 == dispatcher.terminate.call_count

def test_put(obj, workingarea, dispatcher):

    workingarea.put_package.side_effect = [0, 1] # pkgidx
    dispatcher.run.side_effect = [1001, 1002] # runid

    package0 = mock.MagicMock(name='package0')
    package1 = mock.MagicMock(name='package1')

    assert 0 == obj.put(package0)
    assert 1 == obj.put(package1)

    assert [mock.call(package0), mock.call(package1)] == workingarea.put_package.call_args_list
    assert [mock.call(workingarea, 0), mock.call(workingarea, 1)] == dispatcher.run.call_args_list

def test_put_multiple(obj, workingarea, dispatcher):

    workingarea.put_package.side_effect = [0, 1] # pkgidx
    dispatcher.run_multiple.return_value = [1001, 1002] # runid

    package0 = mock.MagicMock(name='package0')
    package1 = mock.MagicMock(name='package1')

    assert [0, 1] == obj.put_multiple([package0, package1])

    assert [mock.call(package0), mock.call(package1)] == workingarea.put_package.call_args_list
    assert [mock.call(workingarea, [0, 1])] == dispatcher.run_multiple.call_args_list

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import io
import pytest

from alphatwirl.collector import WriteListToFile

class MockOpen(object):
    def __init__(self, out):
        self._out = out
    def __call__(self, path):
        return self._out

def mockClose(file):
    pass

##__________________________________________________________________||
@pytest.fixture()
def out():
    return io.BytesIO()

@pytest.fixture()
def obj(out):
    ret = WriteListToFile("tbl.txt")
    ret._open = MockOpen(out)
    ret._close = mockClose
    yield ret

def test_repr(obj):
    repr(obj)

def test_deliver(obj, out):
    results = [
        ('component', 'v1', 'nvar', 'n'),
        ('data1',  100, 6.0,   40),
        ('data1',    2, 9.0, 3.3),
        ('data1', 3124, 3.0, 0.0000001),
        ('data2',  333, 6.0, 300909234),
        ('data2',   11, 2.0, 323432.2234),
    ]

    obj.deliver(results)

    expected = """ component   v1 nvar           n
     data1  100    6          40
     data1    2    9         3.3
     data1 3124    3       1e-07
     data2  333    6   300909234
     data2   11    2 323432.2234
""".encode()

    assert expected == out.getvalue()

def test_deliver_empty_dataframe(obj, out):
    results = [
        ('component', 'v1', 'nvar', 'n'),
    ]

    obj.deliver(results)

    expected = " component v1 nvar n\n".encode()
    assert expected == out.getvalue()

def test_deliver_None_results(obj, out):
    obj.deliver(None)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

from .splitfuncs import create_files_start_length_list

##__________________________________________________________________||
class DatasetIntoEventBuildersSplitter(object):

    def __init__(self, EventBuilder, eventBuilderConfigMaker,
                 maxEvents=-1, maxEventsPerRun=-1,
                 maxFiles=-1, maxFilesPerRun=1
    ):

        self.EventBuilder = EventBuilder
        self.eventBuilderConfigMaker = eventBuilderConfigMaker
        self.max_events = maxEvents
        self.max_events_per_run = maxEventsPerRun
        self.max_files = maxFiles
        self.max_files_per_run = maxFilesPerRun

    def __repr__(self):
        return '{}(EventBuilder={!r}, eventBuilderConfigMaker={!r}, maxEvents={!r}, maxEventsPerRun={!r}, maxFiles={!r}, maxFilesPerRun={!r})'.format(
            self.__class__.__name__,
            self.EventBuilder,
            self.eventBuilderConfigMaker,
            self.max_events,
            self.max_events_per_run,
            self.max_files,
            self.max_files_per_run
        )

    def __call__(self, dataset):

        files = self.eventBuilderConfigMaker.file_list_in(dataset, maxFiles=self.max_files)
        # e.g., ['A.root', 'B.root', 'C.root', 'D.root', 'E.root']

        files_start_length_list = create_files_start_length_list(
            files,
            func_get_nevents_in_file=self.eventBuilderConfigMaker.nevents_in_file,
            max_events=self.max_events,
            max_events_per_run=self.max_events_per_run,
            max_files=self.max_files,
            max_files_per_run=self.max_files_per_run
        )
        # (files, start, length)
        # e.g.,
        # [
        #     (['A.root'], 0, 80),
        #     (['A.root', 'B.root'], 80, 80),
        #     (['B.root'], 60, 80),
        #     (['B.root', 'C.root'], 140, 80),
        #     (['C.root'], 20, 10)
        # ]

        configs = self._create_configs(dataset, files_start_length_list)
        eventBuilders = [self.EventBuilder(c) for c in configs]
        return eventBuilders

    def _create_configs(self, dataset, file_start_length_list):
        configs = [ ]
        for files, start, length in file_start_length_list:
            config = self.eventBuilderConfigMaker.create_config_for(dataset, files, start, length)
            configs.append(config)
        return configs
##__________________________________________________________________||



#!/usr/bin/env python
# Tai Sakuma <sakuma@cern.ch>
import os, sys
import subprocess
import collections
import time
import textwrap
import getpass
import re
import logging

import alphatwirl
from alphatwirl.misc.deprecation import _deprecated_class_method_option

from .exec_util import try_executing_until_succeed, compose_shortened_command_for_logging

##__________________________________________________________________||
# https://htcondor-wiki.cs.wisc.edu/index.cgi/wiki?p=MagicNumbers
HTCONDOR_JOBSTATUS = {
    0: "Unexpanded",
    1: "Idle",
    2: "Running",
    3: "Removed",
    4: "Completed",
    5: "Held",
    6: "Transferring_Output",
    7: "Suspended"
}

##__________________________________________________________________||
## HTCondor Manual:
## 2.5 Submitting a Job
## http://research.cs.wisc.edu/htcondor/manual/v8.4/2_5Submitting_Job.html
##
## condor_submit command manual
## including complete description of submit description file
## http://research.cs.wisc.edu/htcondor/manual/v8.4/condor_submit.html#man-condor-submit

## keys should be in lower case in this dict
DEFAULT_JOB_DESC_DICT = collections.OrderedDict([
    ('executable', 'run.py'),
    ('output', 'results/$(resultdir)/stdout.$(cluster).$(process).txt'),
    ('error', 'results/$(resultdir)/stderr.$(cluster).$(process).txt'),
    ('log', 'results/$(resultdir)/log.$(cluster).$(process).txt'),
    ('arguments', '$(resultdir).p.gz'),
    ('should_transfer_files', 'YES'),
    ('when_to_transfer_output', 'ON_EXIT'),
    ('transfer_input_files', '$(resultdir).p.gz'),
    ('transfer_output_files', 'results'),
    ('universe', 'vanilla'),
    ('notification', 'Error'),
    ('getenv', 'True'),
])

##__________________________________________________________________||
class HTCondorJobSubmitter(object):

    @_deprecated_class_method_option('job_desc_extra', msg='use job_desc_dict instead')
    def __init__(self, job_desc_extra=[ ], job_desc_dict={}):

        self.job_desc_dict = DEFAULT_JOB_DESC_DICT.copy()
        for k, v in job_desc_dict.items():
            self.job_desc_dict[k.lower()] = v # not using update() in case
                                              # job_desc_dict is ordered

        self.user_job_desc_dict = job_desc_dict # for test

        self.job_desc_extra = job_desc_extra # TODO: to be deleted

        self.clusterprocids_outstanding = [ ]
        self.clusterprocids_finished = [ ]

    def run(self, workingArea, package_index):
        return self.run_multiple(workingArea, [package_index])[0]

    def run_multiple(self, workingArea, package_indices):

        if not package_indices:
            return [ ]

        cwd = os.getcwd()
        os.chdir(workingArea.path)

        package_paths = [workingArea.package_path(i) for i in package_indices]
        resultdir_basenames = [os.path.splitext(p)[0] for p in package_paths]
        resultdir_basenames = [os.path.splitext(n)[0] for n in resultdir_basenames]
        resultdirs = [os.path.join('results', n) for n in resultdir_basenames]

        for d in resultdirs:
            alphatwirl.mkdir_p(d)

        self.job_desc_dict['executable'] = workingArea.executable

        extra_input_files = sorted(list(workingArea.extra_input_files))
        if extra_input_files:
            self.job_desc_dict['transfer_input_files'] += ', ' + ', '.join(extra_input_files)

        job_desc = '\n'.join(['{} = {}'.format(k, v) for k, v in self.job_desc_dict.items()])
        job_desc_queue_line = 'queue resultdir in {}'.format(', '.join(resultdir_basenames))

        # TODO: delete this line as job_desc_extra will be obsolete
        job_desc = '\n'.join([job_desc] + self.job_desc_extra)

        job_desc = '\n'.join([job_desc, job_desc_queue_line])

        procargs = ['condor_submit']

        logger = logging.getLogger(__name__)
        command_display = compose_shortened_command_for_logging(procargs)
        logger.debug('execute: {!r}'.format(command_display))

        proc = subprocess.Popen(
            procargs,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        stdout, stderr = proc.communicate(job_desc)
        stdout = stdout.decode()
        stderr = stderr.decode()

        for l in stdout.rstrip().split('\n'):
            logger.debug(l)

        regex = re.compile("(\d+) job\(s\) submitted to cluster (\d+)", re.MULTILINE)
        njobs = int(regex.search(stdout).groups()[0])
        clusterid = regex.search(stdout).groups()[1]
        # e.g., '3158626'

        change_job_priority([clusterid], 10) ## need to make configurable

        procid = ['{}'.format(i) for i in range(njobs)]
        # e.g., ['0', '1', '2', '3']

        clusterprocids = ['{}.{}'.format(clusterid, i) for i in procid]
        # e.g., ['3158626.0', '3158626.1', '3158626.2', '3158626.3']

        self.clusterprocids_outstanding.extend(clusterprocids)

        os.chdir(cwd)

        return clusterprocids

    def poll(self):
        """check if the jobs are running and return a list of cluster IDs for
        finished jobs

        """

        clusterids = clusterprocids2clusterids(self.clusterprocids_outstanding)
        clusterprocid_status_list = query_status_for(clusterids)
        # e.g., [['1730126.0', 2], ['1730127.0', 2], ['1730129.1', 1], ['1730130.0', 1]]


        if clusterprocid_status_list:
            clusterprocids, statuses = zip(*clusterprocid_status_list)
        else:
            clusterprocids, statuses = (), ()

        clusterprocids_finished = [i for i in self.clusterprocids_outstanding if i not in clusterprocids]
        self.clusterprocids_finished.extend(clusterprocids_finished)
        self.clusterprocids_outstanding[:] = clusterprocids

        # logging
        counter = collections.Counter(statuses)
        messages = [ ]
        if counter:
            messages.append(', '.join(['{}: {}'.format(HTCONDOR_JOBSTATUS[k], counter[k]) for k in counter.keys()]))
        if self.clusterprocids_finished:
            messages.append('Finished {}'.format(len(self.clusterprocids_finished)))
        logger = logging.getLogger(__name__)
        logger.info(', '.join(messages))

        return clusterprocids_finished

    def wait(self):
        """wait until all jobs finish and return a list of cluster IDs
        """
        sleep = 5
        while self.clusterprocids_outstanding:
            self.poll()
            time.sleep(sleep)
        return self.clusterprocids_finished

    def failed_runids(self, runids):
        # remove failed clusterprocids from self.clusterprocids_finished
        # so that len(self.clusterprocids_finished)) becomes the number
        # of the successfully finished jobs
        for i in runids:
            try:
                self.clusterprocids_finished.remove(i)
            except ValueError:
                pass

    def terminate(self):
        clusterids = clusterprocids2clusterids(self.clusterprocids_outstanding)
        ids_split = split_ids(clusterids)
        statuses = [ ]
        for ids_sub in ids_split:
            procargs = ['condor_rm'] + ids_sub
            command_display = compose_shortened_command_for_logging(procargs)
            logger = logging.getLogger(__name__)
            logger.debug('execute: {}'.format(command_display))
            proc = subprocess.Popen(
                procargs,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            stdout, stderr = proc.communicate()

##__________________________________________________________________||
def clusterprocids2clusterids(clusterprocids):
    return list(set([i.split('.')[0] for i in clusterprocids]))

##__________________________________________________________________||
def query_status_for(ids, n_at_a_time=500):

    ids_split = split_ids(ids, n=n_at_a_time)
    stdout = [ ]
    for ids_sub in ids_split:
        procargs = ['condor_q'] + ids_sub + ['-format', '%d.', 'ClusterId', '-format', '%d ', 'ProcId', '-format', '%-2s\n', 'JobStatus']
        stdout.extend(try_executing_until_succeed(procargs))

    # e.g., stdout = ['688244.0 1 ', '688245.0 1 ', '688246.0 2 ']

    ret = [l.strip().split() for l in stdout]
    # e.g., [['688244.0', '1'], ['688245.0', '1'], ['688246.0', '2']]

    ret = [[e[0], int(e[1])] for e in ret]
    # a list of [clusterprocid, status]
    # e.g., [['688244.0', 1], ['688245.0', 1], ['688246.0', 2]]

    return ret

##__________________________________________________________________||
def change_job_priority(ids, priority=10, n_at_a_time=500):

    # http://research.cs.wisc.edu/htcondor/manual/v7.8/2_6Managing_Job.html#sec:job-prio

    ids_split = split_ids(ids, n=n_at_a_time)
    for ids_sub in ids_split:
        procargs = ['condor_prio', '-p', str(priority)] + ids_sub
        try_executing_until_succeed(procargs)

##__________________________________________________________________||
def split_ids(ids, n=500):
    # e.g.,
    # ids = [3158174', '3158175', '3158176', '3158177', '3158178']
    # n = 2
    # return [[3158174', '3158175'], ['3158176', '3158177'], ['3158178']]
    return [ids[i:(i + n)] for i in range(0, len(ids), n)]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.datasetloop import DatasetLoop
from alphatwirl.datasetloop import ResumableDatasetLoop

##__________________________________________________________________||
@pytest.fixture()
def datasets():
    dataset1 = mock.Mock()
    dataset2 = mock.Mock()
    return [dataset1, dataset2]

@pytest.fixture()
def reader():
    return mock.Mock()

@pytest.fixture()
def workingarea(monkeypatch):
    module = sys.modules['alphatwirl.datasetloop.loop']
    monkeypatch.setattr(module, 'os', mock.Mock())
    monkeypatch.setattr(module, 'gzip', mock.MagicMock())
    monkeypatch.setattr(module, 'pickle', mock.Mock())
    return mock.Mock()

@pytest.fixture(
    params=[0, 1],
    ids=('DatasetLoop', 'ResumableDatasetLoop')
)
def obj(request, datasets, reader, workingarea):
    if request.param == 0:
        return DatasetLoop(datasets, reader)
    else:
        return ResumableDatasetLoop(datasets, reader, workingarea)

def test_repr(obj):
    repr(obj)

def test_call(obj, reader, datasets):
    result = obj()
    assert [mock.call()] == reader.begin.call_args_list
    assert [mock.call(c) for c in datasets] == reader.read.call_args_list
    assert reader.end() is result

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events
    from alphatwirl.roottree import Branch

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def tree():
    input_file_name = 'sample_03.root'
    tree_name = 'tree'
    input_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', input_file_name)

    input_file = ROOT.TFile.Open(input_path)
    yield input_file.Get(tree_name)

@pytest.fixture()
def events(tree):
    yield Events(tree)

def test_event(events):

    event = events[0]
    assert -125 == event.bChar[0]
    assert 253 == event.bUChar[0]
    assert -10 == event.bShort[0]
    assert 65530 == event.bUShort[0]
    assert -2147483626 == event.bInt[0]
    assert 4294967290 == event.bUInt[0]
    assert -0.123 == pytest.approx(event.bFloat[0], rel=1e-6)
    assert -2.345 == event.bDouble[0]
    assert -4611686018427387900 == event.bLong64[0]
    assert 9223372036854775802 == event.bULong64[0]
    assert 1 == event.bBool[0]

    event = events[1]
    assert 127 == event.bChar[0]
    assert 4 == event.bUChar[0]
    assert 32765 == event.bShort[0]
    assert 8 == event.bUShort[0]
    assert 2147483640 == event.bInt[0]
    assert 12 == event.bUInt[0]
    assert 0.244 == pytest.approx(event.bFloat[0], rel=1e-6)
    assert 3.122 == event.bDouble[0]
    assert 9223372036854775733 == event.bLong64[0]
    assert 123 == event.bULong64[0]
    assert 0 == event.bBool[0]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import array

import pytest

from alphatwirl.roottree import Branch

##__________________________________________________________________||
def test_repr():
    ar = array.array('d', [112.4, 87.6, 30.2])
    ca = array.array('i', [2])
    obj = Branch('jet_pt', ar, ca)
    repr(obj)

def test_array():
    ar = array.array('d', [112.4, 87.6, 30.2])
    ca = array.array('i', [2])
    obj = Branch('jet_pt', ar, ca)

    assert 2 == len(obj)
    assert 112.4 == obj[0]
    assert 87.6 == obj[1]
    with pytest.raises(IndexError):
        obj[2]

def test_value():
    ar = array.array('d', [112.4])
    ca = None
    obj = Branch('met_pt', ar, ca)

    assert 1 == len(obj)
    assert 112.4 == obj[0]
    with pytest.raises(IndexError):
        obj[1]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import numbers
import pytest

from alphatwirl.binning import Round

##__________________________________________________________________||
def test_repr():
    obj = Round()
    repr(obj)

def test_default():
    obj = Round()
    assert 0.5 == obj(0.5)
    assert 0.5 == obj(1.4)
    assert 104.5 == obj(104.5)
    assert -0.5 == obj(-0.4)
    assert -0.5 == obj(-0.5)
    assert -1.5 == obj(-1.4)
    assert -1.5 == obj(-1.5)
    assert -2.5 == obj(-1.6)

def test_width_2():
    obj = Round(2)
    assert -3 == obj( -2.9)
    assert -3 == obj( -2  )
    assert -3 == obj( -1.1)
    assert -1 == obj( -0.9)
    assert -1 == obj(  0  )
    assert -1 == obj(  0.9)
    assert  1 == obj(  1.1)
    assert  1 == obj(  2  )
    assert  1 == obj(  2.9)

def test_width_2_aboundary_0():
    obj = Round(2, 0)
    assert -2 == obj( -1.9)
    assert -2 == obj( -1  )
    assert -2 == obj( -0.1)
    assert  0 == obj(  0.1)
    assert  0 == obj(  1  )
    assert  0 == obj(  1.9)
    assert  2 == obj(  2.1)
    assert  2 == obj(  3  )
    assert  2 == obj(  3.9)

def test_decimal_width():
    obj = Round(0.02, 0.005)
    assert  0.005 == pytest.approx(obj(  0.005))
    assert  0.025 == pytest.approx(obj(  0.025))
    assert  0.065 == pytest.approx(obj(  0.081))
    assert -0.055 == pytest.approx(obj( -0.048))
    assert -0.015 == pytest.approx(obj( -0.015))

def test_keep_aboundary_int():
    # This test is to make sure boundaries are int
    # width/2 is int in both python 2 and 3
    # https://www.python.org/dev/peps/pep-0238/
    obj = Round(4)
    assert 2 == (obj(4))
    assert isinstance(obj(4), numbers.Integral)

def test_on_boundary():
    obj = Round()
    assert  -1.5 == obj( -1.5)
    assert  -0.5 == obj( -0.5)
    assert   0.5 == obj(  0.5)
    assert   1.5 == obj(  1.5)

    obj = Round(0.02, 0.005)
    assert  -0.035 == obj( -0.035)
    assert  -0.015 == obj( -0.015)
    assert   0.005 == obj(  0.005)
    assert   0.025 == obj(  0.025)
    assert   0.045 == obj(  0.045)

def test_next():
    obj = Round()
    assert  -0.5 == obj.next( -1.5)
    assert   0.5 == obj.next( -0.5)
    assert   1.5 == obj.next(  0.5)
    assert   2.5 == obj.next(  1.5)

    obj = Round(0.02, 0.005)
    assert  -0.015 == obj.next( -0.035)
    assert   0.005 == obj.next( -0.015)
    assert   0.025 == obj.next(  0.005)
    assert   0.045 == obj.next(  0.025)
    assert   0.065 == obj.next(  0.045)

def test_valid():
    obj = Round(valid=lambda x: x >= 0)
    assert   0.5 == obj(  1)
    assert  -0.5 == obj(  0)
    assert  None == obj( -1)

def test_min_int():
    obj = Round(10, 100, min=30)
    assert   100 == obj( 100)
    assert    30 == obj(  30) # this works for int
    assert  None == obj(  29)

def test_min_float_not_a_boundary():
    obj = Round(0.2, 2.0, min=1.1)
    # boundaries = [0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]
    # min=1.1 is not a boundary

    assert 2.0 == obj(2.0)
    # this is always true because 2.0 is the given boundary.

    assert 1.0 == pytest.approx(obj(1.05))
    # 1.05 is below min=1.1. However, it is above the lower edge of
    # the bin 1.1 belongs, which is 1.0. So obj(1.05) should returns
    # the lower edge.

    assert obj(0.95) is None

@pytest.mark.parametrize('width', [0.1, 0.2])
@pytest.mark.parametrize('underflow_bin', [None, -1, 0, 0.1])
def test_min_on_a_boundary(width, underflow_bin):
    # this test is related to
    # the issue 43
    # https://github.com/alphatwirl/alphatwirl/issues/43

    min_ = 1.0 # on a boundary (within rounding of float)
    obj = Round(width, 2.0, min=min_, underflow_bin=underflow_bin)

    # example boundaries (depending on the architecture)
    # when width = 0.1:
    #     boundaries = [
    #         0.9999999999999992, 1.0999999999999992,
    #         1.1999999999999993, 1.2999999999999994, 1.3999999999999995,
    #         1.4999999999999996, 1.5999999999999996, 1.6999999999999997,
    #         1.7999999999999998, 1.9, 2.0]
    #   min=1.0 is the 1st element. but the precise value is slightly
    #   less than 1.0
    #
    # when width = 0.2:
    #     boundaries = [
    #         0.8000000000000003, 1.0000000000000002,
    #         1.2000000000000002, 1.4000000000000001, 1.6, 1.8, 2.0]
    #   min=1.0 is the 2nd element, slightly greater than 1.0.

    assert 2.0 == obj(2.0)
    # this is always true because 2.0 is the given boundary.

    # min=1.0 is a boundary, but not necessarily exact.
    if min_ == pytest.approx(obj(min_)):
        assert obj(1.05) == obj(min_)
        assert obj(0.95) is underflow_bin
    else:
        assert obj(0.95) == obj(min_)
        assert obj(0.95) is not underflow_bin
    # the results depend on the architecture.
    # it is wise to not set min a boundary.

    assert obj(0.7) is underflow_bin

    if underflow_bin is None:
        assert obj.next(underflow_bin) is None
    else:
        assert obj(min_) == obj.next(underflow_bin)
        # the next to the underflow bin is the bin for the min

def test_max_int():
    obj = Round(10, 100, max=150)
    assert   100 == obj( 100)
    assert  None == obj( 150)
    assert  None == obj( 500)

    assert  None == obj.next(140) # the next to the last bin is
                                  # the overflow bin

@pytest.mark.parametrize('width', [0.1, 0.2])
@pytest.mark.parametrize('max_', [1.6, 2.4])
@pytest.mark.parametrize('overflow_bin', [None, True, 3.0])
def test_max_float_on_a_boundary(width, max_, overflow_bin):

    # max_ on a boundary (within rounding of float)
    obj = Round(width, 1.0, max=max_, overflow_bin=overflow_bin)

    # example boundaries (depending on the architecture)
    # when width = 0.1:
    #     boundaries = [
    #         1.0, 1.1, 1.2000000000000002, 1.3000000000000003,
    #         1.4000000000000004, 1.5000000000000004, 1.6000000000000005,
    #         1.7000000000000006, 1.8000000000000007, 1.9000000000000008,
    #         2.000000000000001, 2.100000000000001, 2.200000000000001,
    #         2.300000000000001, 2.4000000000000012]
    #   slightly greater than 1.6 and 2.4
    #
    # when width = 0.2:
    #     boundaries = [
    #         1.0, 1.2, 1.4, 1.5999999999999999, 1.7999999999999998,
    #         1.9999999999999998, 2.1999999999999997, 2.4]
    #   slightly less than 1.6, exactly 2.4!

    #
    if overflow_bin is True:
        overflow_bin = obj.boundaries[-1]

    assert 1.0 == obj(1.0)
    # this is always true because 1.0 is the given boundary.

    # If the max is exactly a boundary, the max is the upper edge of
    # the last bin and bin(max) is overflow because the upper edge is
    # open. Otherwise, bin(max) is in the last bin.
    if max_ == obj.boundaries[-1]:
        assert obj(max_) == overflow_bin
    else:
        assert obj.boundaries[-2] == obj(max_)

    # The max is either the upper or lower edge of the last bin, but
    # not necessarily exact.
    if not max_ == pytest.approx(obj.boundaries[-1]):
        assert max_ == pytest.approx(obj.boundaries[-2])

    # the next to the last bin is the overflow bin
    assert obj.next(obj.boundaries[-2]) is overflow_bin

    # the next to the overflow bin is the overflow bin
    assert obj.next(overflow_bin) is overflow_bin

@pytest.mark.parametrize('overflow_bin', [None, True, 150, 300])
def test_max_int_on_a_boundary(overflow_bin):
    # max is always exactly a boundary
    max_ = 150
    obj = Round(10, 100, max=max_, overflow_bin=overflow_bin)
    assert 100 == obj(100)

    if overflow_bin is True:
        overflow_bin = obj.boundaries[-1]

    assert max_ == obj.boundaries[-1]
    assert obj(max_) is overflow_bin
    assert obj.next(obj.boundaries[-2]) is overflow_bin
    assert obj.next(overflow_bin) is overflow_bin

def test_inf():
    obj = Round(10, 100)
    assert obj(float('inf')) is None
    assert obj(float('-inf')) is None
    assert obj.next(float('inf')) is None
    assert obj.next(float('-inf')) is None

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import EventsInDatasetReader, EventLoop

##__________________________________________________________________||
@pytest.fixture()
def eventLoopRunner():
    return mock.Mock(name='eventLoopRunner')

@pytest.fixture()
def reader():
    ret = mock.Mock(name='reader')
    ret.configure_mock(name='reader')
    return ret

@pytest.fixture()
def collector():
    return mock.Mock(name='collector')

@pytest.fixture()
def split_into_build_events():
    return mock.Mock(name='split_into_build_events')

@pytest.fixture()
def obj(eventLoopRunner, reader, collector, split_into_build_events):
    return EventsInDatasetReader(eventLoopRunner, reader, collector,
                                 split_into_build_events)

##__________________________________________________________________||
def test_deprecated(eventLoopRunner, reader, collector, split_into_build_events, caplog):
    with caplog.at_level(logging.WARNING):
        obj = EventsInDatasetReader(
            eventLoopRunner, reader, collector,
            split_into_build_events
        )

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'EventsInDatasetReader' in caplog.records[0].name
    assert 'deprecated' in caplog.records[0].msg

##__________________________________________________________________||
def test_repr(obj):
   repr(obj)

def test_begin(obj, eventLoopRunner):
    assert 0 == eventLoopRunner.begin.call_count
    obj.begin()
    assert 1 == eventLoopRunner.begin.call_count

def test_begin_end(obj, eventLoopRunner, collector):
    obj.begin()
    eventLoopRunner.end.return_value = [ ]
    end = obj.end()
    assert [mock.call([ ])] == collector.collect.call_args_list
    assert collector.collect() == end

def test_standard(obj, eventLoopRunner, reader, collector,
                  split_into_build_events, caplog):

    ## begin
    obj.begin()

    ## create data sets
    # dataset1 - 3 event builders
    build_events1 = mock.Mock(name='build_events1')
    build_events2 = mock.Mock(name='build_events2')
    build_events3 = mock.Mock(name='build_events3')
    dataset1 = mock.Mock(name='dataset1', build_events=[build_events1, build_events2, build_events3])
    dataset1.configure_mock(name='dataset1')

    # dataset2 - no event builder
    dataset2 = mock.Mock(name='dataset2', build_events=[ ])
    dataset2.configure_mock(name='dataset2')

    # dataset3 - 1 event builder
    build_events4 = mock.Mock(name='build_events4')
    dataset3 = mock.Mock(name='dataset3', build_events=[build_events4])
    dataset3.configure_mock(name='dataset3')

    split_into_build_events.side_effect = lambda dataset: dataset.build_events

    ## read
    obj.read(dataset1)
    obj.read(dataset2)
    obj.read(dataset3)

    assert 0 == eventLoopRunner.run.call_count
    assert 3 == eventLoopRunner.run_multiple.call_count

    call1 = eventLoopRunner.run_multiple.call_args_list[0]

    eventLoop1 = call1[0][0][0]
    assert isinstance(eventLoop1, EventLoop)
    assert build_events1 is eventLoop1.build_events
    assert reader is not eventLoop1.reader
    assert 'reader' == eventLoop1.reader.name

    eventLoop2 = call1[0][0][1]
    assert isinstance(eventLoop2, EventLoop)
    assert build_events2 is eventLoop2.build_events
    assert reader is not eventLoop2.reader
    assert 'reader' == eventLoop2.reader.name

    eventLoop3 = call1[0][0][2]
    assert isinstance(eventLoop3, EventLoop)
    assert build_events3 is eventLoop3.build_events
    assert reader is not eventLoop3.reader
    assert 'reader' == eventLoop3.reader.name

    call2 = eventLoopRunner.run_multiple.call_args_list[1]
    assert mock.call([ ]) == call2

    call3 = eventLoopRunner.run_multiple.call_args_list[2]

    eventLoop4 = call3[0][0][0]
    assert isinstance(eventLoop4, EventLoop)
    assert build_events4 is eventLoop4.build_events
    assert reader is not eventLoop4.reader
    assert 'reader' == eventLoop4.reader.name

    ## end
    eventLoopRunner.end.return_value = [
        eventLoop1.reader, eventLoop2.reader, eventLoop3.reader, eventLoop4.reader
    ]
    collector.collect.side_effect = lambda x: x
    assert 0 == eventLoopRunner.end.call_count
    assert 0 == collector.collect.call_count
    results = obj.end()
    assert 1 == eventLoopRunner.end.call_count
    assert 1 == collector.collect.call_count
    expected = [
        ('dataset1', (eventLoop1.reader, eventLoop2.reader, eventLoop3.reader), ),
        ('dataset2', ( )),
        ('dataset3', (eventLoop4.reader, )),
    ]
    assert expected == results
    assert [mock.call(expected)] == collector.collect.call_args_list

def test_wrong_number_of_results(obj, eventLoopRunner, reader,
                                 collector, split_into_build_events,
                                 caplog):

    reader1 = mock.Mock(name='reader1')
    reader2 = mock.Mock(name='reader2')
    eventLoopRunner.end.return_value = [reader1, reader2]

    dataset1 = mock.Mock(name='dataset1')
    obj.dataset_nreaders[:] = [(dataset1, 1)]

    with caplog.at_level(logging.WARNING, logger = 'alphatwirl'):
        results = obj.end()

    assert results is None

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'EventsInDatasetReader' in caplog.records[0].name
    assert 'the same number of' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from ..summary import Reader, Summarizer, NextKeyComposer, KeyValueComposer
from ..collector import ToTupleListWithDatasetColumn
from ..collector import WriteListToFile
from ..loop import Collector

##__________________________________________________________________||
def build_counter_collector_pair(tblcfg):
    keyValComposer = KeyValueComposer(
        keyAttrNames=tblcfg['keyAttrNames'],
        binnings=tblcfg['binnings'],
        keyIndices=tblcfg['keyIndices'],
        valAttrNames=tblcfg['valAttrNames'],
        valIndices=tblcfg['valIndices']
    )
    nextKeyComposer = NextKeyComposer(tblcfg['binnings']) if tblcfg['binnings'] is not None else None
    summarizer = Summarizer(
        Summary=tblcfg['summaryClass']
    )
    reader = Reader(
        keyValComposer=keyValComposer,
        summarizer=summarizer,
        nextKeyComposer=nextKeyComposer,
        weightCalculator=tblcfg['weight'],
        nevents=tblcfg['nevents']
    )
    resultsCombinationMethod = ToTupleListWithDatasetColumn(
        summaryColumnNames = tblcfg['keyOutColumnNames'] + tblcfg['valOutColumnNames']
    )
    deliveryMethod = WriteListToFile(tblcfg['outFilePath']) if tblcfg['outFile'] else None
    collector = Collector(resultsCombinationMethod, deliveryMethod)
    return reader, collector

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules.Count import Count

##__________________________________________________________________||
@pytest.fixture()
def obj1_results_org():
    return [
        [1, 'class1', 'name1', 2, 2],
        [1, 'class1', 'name2', 1, 2],
        [1, 'class2', 'name3', 0, 1],
    ]

@pytest.fixture()
def obj1(obj1_results_org):
    ret = Count()
    ret._results = copy.deepcopy(obj1_results_org)
    return ret

@pytest.fixture()
def obj2_results_org():
    return [
        [1, 'class1', 'name1', 3, 5],
        [1, 'class1', 'name2', 2, 4],
        [1, 'class2', 'name3', 1, 2],
    ]

@pytest.fixture()
def obj2(obj2_results_org):
    ret = Count()
    ret._results = copy.deepcopy(obj2_results_org)
    return ret

@pytest.fixture()
def expected():
    return [
        [1, 'class1', 'name1', 5, 7],
        [1, 'class1', 'name2', 3, 6],
        [1, 'class2', 'name3', 1, 3],
    ]

def test_add(obj1, obj2, obj1_results_org, obj2_results_org, expected):
    obj3 = obj1 + obj2
    assert expected == obj3._results
    assert obj1._results is not obj3._results
    assert obj2._results is not obj3._results

    assert obj1_results_org is not obj1._results
    assert obj1_results_org == obj1._results

    assert obj2_results_org is not obj2._results
    assert obj2_results_org == obj2._results

def test_radd(obj1, obj2, obj1_results_org, obj2_results_org, expected):
    obj3 = sum([obj1, obj2]) # 0 + obj1 is executed
    assert expected == obj3._results
    assert obj1._results is not obj3._results
    assert obj2._results is not obj3._results

    assert obj1_results_org is not obj1._results
    assert obj1_results_org == obj1._results

    assert obj2_results_org is not obj2._results
    assert obj2_results_org == obj2._results

def test_iadd(obj1, obj2, obj1_results_org, obj2_results_org, expected):
    obj3 = obj1
    obj3 += obj2

    assert obj3 is obj1
    assert expected == obj3._results

    assert obj1_results_org is not obj1._results

    assert obj2_results_org is not obj2._results
    assert obj2_results_org == obj2._results

##__________________________________________________________________||
def test_add_incompatible_different_length(obj1, caplog):
    obj2 = Count()
    obj2._results  = [
        [1, 'class1', 'name1', 3, 5],
        [1, 'class1', 'name2', 2, 4],
    ]

    with caplog.at_level(logging.WARNING):
        obj3 = obj1 + obj2

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'Count' in caplog.records[0].name
    assert 'cannot add' in caplog.records[0].msg

    assert obj1._results == obj3._results

def test_add_incompatible_different_first_values(obj1, caplog):
    obj2 = Count()
    obj2._results  = [
        [1, 'class1', 'name1', 3, 5],
        [1, 'class1', 'name2', 2, 4],
        [1, 'class2', 'name4', 1, 2],
    ]

    with caplog.at_level(logging.WARNING):
        obj3 = obj1 + obj2

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'Count' in caplog.records[0].name
    assert 'cannot add' in caplog.records[0].msg

    obj3 = obj1 + obj2
    assert obj1._results == obj3._results

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
from __future__ import print_function
import os
import errno
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl import mkdir_p

##__________________________________________________________________||
@pytest.fixture()
def mock_makedirs(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(os, 'makedirs', ret)
    return ret

@pytest.fixture()
def mock_isdir(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(os.path, 'isdir', ret)
    return ret

##__________________________________________________________________||
def test_emtpy(mock_makedirs):
    mkdir_p('')
    assert [ ] == mock_makedirs.call_args_list

def test_success(mock_makedirs):
    mkdir_p('a/b')
    assert [mock.call('a/b')] == mock_makedirs.call_args_list

def test_already_exist(mock_makedirs, mock_isdir, caplog):
    mock_isdir.return_value = True
    mock_makedirs.side_effect = OSError(errno.EEXIST, 'already exist')
    with caplog.at_level(logging.DEBUG - 1):
        mkdir_p('a/b')

    assert [mock.call('a/b')] == mock_makedirs.call_args_list
    assert len(caplog.records) == 1
    assert caplog.records[0].levelno == logging.DEBUG - 1
    assert 'tried' in caplog.records[0].msg

def test_raise(mock_makedirs, mock_isdir, caplog):
    mock_isdir.return_value = False
    mock_makedirs.side_effect = OSError
    with pytest.raises(OSError):
        mkdir_p('a/b')

    assert [mock.call('a/b')] == mock_makedirs.call_args_list

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.selection.factories.expand import expand_path_cfg

##__________________________________________________________________||
@pytest.fixture()
def alias_dict():
    return {
        'var_cut': ('ev : {low} <= ev.var[0] < {high}', dict(low=10, high=200)),
    }

##__________________________________________________________________||
params = [
    pytest.param(
        ('var_cut', dict(name='var_cut25', low=25)),
        dict(
            factory='LambdaStrFactory',
            components=(),
            lambda_str='ev : {low} <= ev.var[0] < {high}',
            name='var_cut25',
            low=25, high=200,
        ),
        id='example1'
    ),
]

@pytest.mark.parametrize('path_cfg, expected', params)
def test_alias(alias_dict, path_cfg, expected):
    actual = expand_path_cfg(path_cfg=path_cfg, alias_dict=alias_dict)
    assert expected == actual

    # give expanded one
    actual = expand_path_cfg(path_cfg=actual, alias_dict=alias_dict)
    assert expected == actual


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import gzip

try:
   import cPickle as pickle
except:
   import pickle

import pytest

import alphatwirl

##__________________________________________________________________||
@pytest.fixture()
def env():
   # add path to alphatwirl to PYTHONPATH in env for subprocess
   env = os.environ.copy()
   alphatwirl_path = os.path.dirname(os.path.dirname(os.path.abspath(alphatwirl.__file__)))
   if 'PYTHONPATH' in env:
      env['PYTHONPATH'] = '{}:{}'.format(alphatwirl_path, env['PYTHONPATH'])
   else:
      env['PYTHONPATH'] = alphatwirl_path
   return env

@pytest.fixture()
def workingarea_path(tmpdir_factory):
   ret = str(tmpdir_factory.mktemp(''))
   return ret

##__________________________________________________________________||
class Reader(object):
   def end(self):
      pass

@pytest.fixture()
def pickled_reader_path(workingarea_path):

   reader = Reader()

   ret = os.path.join(workingarea_path, 'reader.p.gz')
   with gzip.open(ret, 'wb') as f:
      pickle.dump(reader, f, protocol = pickle.HIGHEST_PROTOCOL)

   return ret

@pytest.mark.script_launch_mode('subprocess')
def test_run(script_runner, pickled_reader_path, env):

   script_path = os.path.join(os.path.dirname(alphatwirl.datasetloop.__file__), 'resume.py')
   args = [pickled_reader_path]
   ret = script_runner.run(script_path, *args, env=env)
   assert ret.success
   assert '' == ret.stdout

   # assert '' == ret.stderr
   ## commented out because of "RuntimeWarning" in travis
   ## https://travis-ci.org/alphatwirl/alphatwirl/jobs/409237073
   ## E AssertionError: assert '' == '/home/travis/miniconda/envs...ok( name,
   ## *args, **kwds )\n'
   ## E + /home/travis/miniconda/envs/testenv/lib/ROOT.py:301: RuntimeWarning:
   ## numpy.dtype size changed, may indicate binary incompatibility. Expected
   ## 96, got 88


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import copy
import numpy as np
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.summary import Summarizer
from alphatwirl.summary import Sum

##__________________________________________________________________||
@pytest.fixture()
def obj1():
    ret = Summarizer(Summary=Sum)
    ret._results.update({
        (1, ): Sum(contents=np.array((4, ))),
        (2, ): Sum(contents=np.array((3, ))),
        (3, ): Sum(contents=np.array((2, ))),
    })
    return ret

@pytest.fixture()
def obj2():
    ret = Summarizer(Summary=Sum)
    ret._results.update({
        (2, ): Sum(contents=np.array((3.2, ))),
        (4, ): Sum(contents=np.array((2, ))),
    })
    return ret

@pytest.fixture()
def expected():
    return {
        (1, ): Sum(contents=np.array((4, ))),
        (2, ): Sum(contents=np.array((6.2, ))),
        (3, ): Sum(contents=np.array((2, ))),
        (4, ): Sum(contents=np.array((2, ))),
    }

def test_add(obj1, obj2, expected):
    obj3 = obj1 + obj2
    assert expected == obj3._results
    assert obj1._results[(1, )] is not obj3._results[(1, )]
    assert obj2._results[(4, )] is not obj3._results[(4, )]

def test_radd(obj1, obj2, expected):
    obj3 = sum([obj1, obj2]) # 0 + obj1 is executed
    assert expected == obj3._results
    assert obj1._results[(1, )] is not obj3._results[(1, )]
    assert obj2._results[(4, )] is not obj3._results[(4, )]

def test_iadd(obj1, obj2, expected):
    obj3 = obj1

    obj3 += obj2
    assert obj3 is obj1
    assert expected == obj3._results

def test_copy(obj1):
    copy1 = copy.copy(obj1)
    assert obj1._results == copy1._results
    assert obj1._results[(1, )]is not copy1._results[(1, )]
    assert obj1._results[(2, )]is not copy1._results[(2, )]
    assert obj1._results[(3, )]is not copy1._results[(3, )]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import numbers
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import inspect_tree

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def tree():
    input_file_name = 'sample_04.root'
    tree_name = 'tree'
    input_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', input_file_name)

    input_file = ROOT.TFile.Open(input_path)
    yield input_file.Get(tree_name)


def test_inspect(tree):
    actual = inspect_tree(tree)

    ## assert expected == actual ## this doesn't work in different
                                 ## architectures.

    assert len(expected) == len(actual) # only 'leaves'
    leaves_expected = expected['leaves']
    leaves_actual = actual['leaves']
    assert len(leaves_expected) == len(leaves_actual)
    for leave_expected, leave_actual in zip(leaves_expected, leaves_actual):
        assert len(leave_expected) == len(leave_actual)
        keys = list(leave_expected.keys())
        for key in keys:
            e = leave_expected[key]
            a = leave_actual[key]
            if isinstance(e, numbers.Integral):
                assert e == a
            elif isinstance(e, numbers.Number):
                assert e == pytest.approx(a, rel=1e-1)
            else:
                assert e == a

expected = {
    'leaves': [
        {'compression_factor': 7.413333333333333,
         'countname': None,
         'isarray': '0',
         'name': 'nvar',
         'size': 7.152557373046875e-05,
         'title': 'nvar/I',
         'type': 'Int_t',
         'uncompressed_size': 0.000530242919921875},
        {'compression_factor': 7.5813953488372094,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bChar',
         'size': 8.20159912109375e-05,
         'title': 'bChar[nvar]/B',
         'type': 'Char_t',
         'uncompressed_size': 0.000621795654296875},
        {'compression_factor': 7.551724137931035,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bUChar',
         'size': 8.296966552734375e-05,
         'title': 'bUChar[nvar]/b',
         'type': 'UChar_t',
         'uncompressed_size': 0.0006265640258789062},
        {'compression_factor': 7.426966292134831,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bShort',
         'size': 8.487701416015625e-05,
         'title': 'bShort[nvar]/S',
         'type': 'Short_t',
         'uncompressed_size': 0.0006303787231445312},
        {'compression_factor': 7.4,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bUShort',
         'size': 8.58306884765625e-05,
         'title': 'bUShort[nvar]/s',
         'type': 'UShort_t',
         'uncompressed_size': 0.0006351470947265625},
        {'compression_factor': 7.164835164835165,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bInt',
         'size': 8.678436279296875e-05,
         'title': 'bInt[nvar]/I',
         'type': 'Int_t',
         'uncompressed_size': 0.000621795654296875},
        {'compression_factor': 7.141304347826087,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bUInt',
         'size': 8.7738037109375e-05,
         'title': 'bUInt[nvar]/i',
         'type': 'UInt_t',
         'uncompressed_size': 0.0006265640258789062},
        {'compression_factor': 7.193548387096774,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bFloat',
         'size': 8.869171142578125e-05,
         'title': 'bFloat[nvar]/F',
         'type': 'Float_t',
         'uncompressed_size': 0.0006380081176757812},
        {'compression_factor': 6.764705882352941,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bDouble',
         'size': 9.72747802734375e-05,
         'title': 'bDouble[nvar]/D',
         'type': 'Double_t',
         'uncompressed_size': 0.0006580352783203125},
        {'compression_factor': 6.764705882352941,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bLong64',
         'size': 9.72747802734375e-05,
         'title': 'bLong64[nvar]/L',
         'type': 'Long64_t',
         'uncompressed_size': 0.0006580352783203125},
        {'compression_factor': 6.747572815533981,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bULong64',
         'size': 9.822845458984375e-05,
         'title': 'bULong64[nvar]/l',
         'type': 'ULong64_t',
         'uncompressed_size': 0.0006628036499023438},
        {'compression_factor': 7.5813953488372094,
         'countname': 'nvar',
         'isarray': '1',
         'name': 'bBool',
         'size': 8.20159912109375e-05,
         'title': 'bBool[nvar]/O',
         'type': 'Bool_t',
         'uncompressed_size': 0.000621795654296875}]}

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently.exec_util import try_executing_until_succeed, compose_shortened_command_for_logging

##__________________________________________________________________||
def test_without_monkeypatch_subproces(caplog):
    procargs = ['ls']
    with caplog.at_level(logging.DEBUG, logger = 'alphatwirl'):
        try_executing_until_succeed(procargs)

##__________________________________________________________________||
@pytest.fixture()
def subprocess(monkeypatch):
    module = sys.modules['alphatwirl.concurrently.exec_util']
    ret = mock.MagicMock(name='subprocess')
    monkeypatch.setattr(module, 'subprocess', ret)
    return ret

def test_success(subprocess):
    procargs = ['ls']

    proc = mock.MagicMock(name='ls')
    proc.communicate.return_value = (b'aaa bbb', b'')
    proc.returncode = 0
    subprocess.Popen.side_effect = [proc]

    assert [b'aaa bbb'] == try_executing_until_succeed(procargs)

def test_fail_success(subprocess, caplog):
    procargs = ['ls']

    proc0 = mock.MagicMock(name='ls')
    proc0.communicate.return_value = (b'', b'')
    proc0.returncode = 1

    proc1 = mock.MagicMock(name='ls')
    proc1.communicate.return_value = (b'aaa bbb', b'')
    proc1.returncode = 0

    subprocess.Popen.side_effect = [proc0, proc1]

    with caplog.at_level(logging.WARNING, logger = 'alphatwirl'):
        assert [b'aaa bbb'] == try_executing_until_succeed(procargs, sleep=0.1)

##__________________________________________________________________||
def test_compose_shortened_command_for_logging():
    procargs = ['condor_q', '3158110', '3158111', '3158112', '3158113',
                '3158114', '3158115', '3158116', '3158117', '3158118', '3158119',
                '3158120', '3158121', '3158122', '3158123', '3158124', '3158125',
                '-format', '%-2s ', 'ClusterId', '-format', '%-2s\n', 'JobStatus']
    expected = 'condor_q 3158110 3158111 3158112 3158113 3158114 3...((86 letters))... -format "%-2s " ClusterId -format %-2s\n JobStatus'
    actual = compose_shortened_command_for_logging(procargs)
    assert expected == actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import itertools

##__________________________________________________________________||
def key_vals_dict_to_tuple_list(key_vals_dict, fill=float('nan')):
    """Convert ``key_vals_dict`` to `tuple_list``.

    Args:
        key_vals_dict (dict): The first parameter.
        fill: a value to fill missing data

    Returns:
        A list of tuples

    """

    tuple_list = [ ]

    if not key_vals_dict: return tuple_list

    vlen = max([len(vs) for vs in itertools.chain(*key_vals_dict.values())])

    for k, vs in key_vals_dict.items():
        try:
            tuple_list.extend([k + tuple(v) + (fill, )*(vlen - len(v)) for v in vs])
        except TypeError:
            # assume k is not a tuple
            tuple_list.extend([(k, ) + tuple(v) + (fill, )*(vlen - len(v)) for v in vs])


    return tuple_list

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

from alphatwirl.binning import Binning

##__________________________________________________________________||
def test_repr():
    boundaries = (10, 20, 30, 40, 50)
    obj = Binning(boundaries=boundaries)
    repr(obj)

def test_call():
    bins = (1, 2, 3, 4)
    lows = (10.0, 20.0, 30.0, 40.0)
    ups = (20.0, 30.0, 40.0, 50.0)
    obj = Binning(bins=bins, lows=lows, ups=ups, retvalue='number')
    assert 1 == obj(15)
    assert 2 == obj(21)
    assert 2 == obj(20) # on the low edge
    assert 0 == obj(5) # underflow
    assert 5 == obj(55) # overflow

def test_onBoundary():
    boundaries = (0.000001, 0.00001, 0.0001)
    obj = Binning(boundaries=boundaries, retvalue='number')
    assert  1 == obj( 0.000001 )
    assert  2 == obj( 0.00001  )
    assert  3 == obj( 0.0001   )

def test_lowedge():
    lows = (10.0, 20.0, 30.0, 40.0)
    ups = (20.0, 30.0, 40.0, 50.0)
    obj = Binning(lows=lows, ups=ups, retvalue='lowedge')
    assert             10 == obj( 15 )
    assert             20 == obj( 21 )
    assert             20 == obj( 20 )
    assert  float("-inf") == obj(  5 )
    assert             50 == obj( 55 )

    obj = Binning(lows=lows, ups=ups) # 'lowedge' is default
    assert             10 == obj( 15 )
    assert             20 == obj( 21 )
    assert             20 == obj( 20 )
    assert  float("-inf") == obj(  5 )
    assert             50 == obj( 55 )

def test_init_retvalue():
    boundaries = (10, 20, 30, 40, 50)
    Binning(boundaries=boundaries)
    Binning(boundaries=boundaries, retvalue='number')
    Binning(boundaries=boundaries, retvalue='lowedge')
    with pytest.raises(ValueError):
        Binning(boundaries=boundaries, retvalue='center')

    with pytest.raises(ValueError):
        Binning(boundaries=boundaries, retvalue='yyy')

    with pytest.raises(ValueError):
        Binning(boundaries=boundaries, retvalue='lowedge', bins=(1, 2, 3, 4))

    with pytest.raises(ValueError):
        Binning(boundaries=boundaries, retvalue='lowedge', underflow_bin=-1)

    with pytest.raises(ValueError):
        Binning(boundaries=boundaries, retvalue='lowedge', overflow_bin=-1)

def test_init_with_bins_lows_ups():
    bins = (1, 2, 3, 4)
    lows = (10.0, 20.0, 30.0, 40.0)
    ups = (20.0, 30.0, 40.0, 50.0)
    boundaries = (10, 20, 30, 40, 50)
    obj = Binning(bins=bins, lows=lows, ups=ups, retvalue='number')
    assert bins == obj.bins
    assert boundaries == obj.boundaries

def test_init_with_lows_ups():
    bins = (1, 2, 3, 4)
    lows = (10.0, 20.0, 30.0, 40.0)
    ups = (20.0, 30.0, 40.0, 50.0)
    boundaries = (10, 20, 30, 40, 50)
    obj = Binning(lows=lows, ups=ups, retvalue='number')
    assert bins == obj.bins
    assert boundaries == obj.boundaries

def test_init_with_boundaries():
    bins = (1, 2, 3, 4)
    lows = (10.0, 20.0, 30.0, 40.0)
    ups = (20.0, 30.0, 40.0, 50.0)
    boundaries = (10, 20, 30, 40, 50)
    obj = Binning(boundaries=boundaries, retvalue='number')
    assert bins == obj.bins
    assert lows == obj.lows
    assert ups == obj.ups

def test_init_exceptions():
    with pytest.raises(ValueError):
        Binning()

    with pytest.raises(ValueError):
        Binning(lows=1)

    with pytest.raises(ValueError):
        Binning(ups=1)

    with pytest.raises(ValueError):
        Binning(boundaries=1, lows=1)

    with pytest.raises(ValueError):
        Binning(boundaries=1, ups=1)

    with pytest.raises(ValueError):
        Binning(boundaries=1, lows=1, ups=1)

    lows = (10.0, 20.0, 30.0, 45.0)
    ups = (20.0, 30.0, 40.0, 50.0)

    with pytest.raises(ValueError):
        Binning(lows=lows, ups=ups)

def test_init_exceptions_nobin():
    boundaries = (10, )

    with pytest.raises(ValueError):
        Binning(boundaries=boundaries)

def test_next_number():
    boundaries = (10, 20, 30, 40, 50)
    obj = Binning(boundaries=boundaries, retvalue='number')
    assert  1 == obj.next(0)
    assert  2 == obj.next(1)
    assert  3 == obj.next(2)
    assert  4 == obj.next(3)
    assert  5 == obj.next(4)
    assert  5 == obj.next(5)

    assert 5 == obj.next(5) # overflow_bin returns the same

    with pytest.raises(ValueError):
         obj.next(2.5)

    with pytest.raises(ValueError):
        obj.next(6)

def test_next_lowedge():
    boundaries = (10, 20, 30, 40, 50)
    obj = Binning(boundaries=boundaries, retvalue='lowedge')

    # on the boundaries
    assert  20 == obj.next(10)
    assert  30 == obj.next(20)
    assert  40 == obj.next(30)
    assert  50 == obj.next(40)
    assert  50 == obj.next(50)

    # underflow_bin
    assert 10 == obj.next(float('-inf'))

    boundaries = (0.001, 0.002, 0.003, 0.004, 0.005)
    obj = Binning(boundaries=boundaries, retvalue='lowedge')
    assert  0.002 == obj.next( 0.001)
    assert  0.003 == obj.next( 0.002)
    assert  0.004 == obj.next( 0.003)
    assert  0.005 == obj.next( 0.004)
    assert  0.005 == obj.next( 0.005)

def test_valid():
    obj = Binning(boundaries=(30, 40, 50), retvalue='number', valid=lambda x: x >= 10)
    assert  1 == obj( 33)
    assert  2 == obj( 45)
    assert obj( 9) is None

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
import numpy as np
import copy

##__________________________________________________________________||
class Scan(object):
    def __init__(self, val = None, weight = 1, contents = None):

        if contents is not None:
            self.contents = contents
            return

        if val is None:
            self.contents = [ ]
            return

        self.contents = [val]

    def __add__(self, other):
        contents = self.contents + other.contents
        return self.__class__(contents = contents)

    def __radd__(self, other):
        # is called with other = 0 when e.g. sum([obj1, obj2])
        if other == 0:
            return self.__class__() + self
        raise TypeError('unsupported: {!r} + {!r}'.format(other, self))

    def __repr__(self):
        return '{}(contents = {})'.format(self.__class__.__name__, self.contents)

    def __eq__(self, other):
        return self.contents == other.contents

    def __copy__(self):
        contents = list(self.contents)
        return self.__class__(contents = contents)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import logging
import textwrap
import collections

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import WorkingArea
from alphatwirl.concurrently import HTCondorJobSubmitter

##__________________________________________________________________||
@pytest.fixture()
def proc_submit():
    ret =  mock.MagicMock(name='proc_condor_submit')
    ret.communicate.return_value = (b'1 job(s) submitted to cluster 1012.', b'')
    return ret

@pytest.fixture()
def proc_prio():
    ret = mock.MagicMock(name='proc_condor_prio')
    ret.communicate.return_value = ('', '')
    ret.returncode = 0
    return ret

@pytest.fixture()
def mocksubprocess(monkeypatch, proc_submit, proc_prio):
    ret = mock.MagicMock(name='subprocess')

    module = sys.modules['alphatwirl.concurrently.HTCondorJobSubmitter']
    monkeypatch.setattr(module, 'subprocess', ret)

    module = sys.modules['alphatwirl.concurrently.exec_util']
    monkeypatch.setattr(module, 'subprocess', ret)

    ret.Popen.side_effect = [proc_submit, proc_prio]
    return ret

@pytest.fixture()
def obj(mocksubprocess):
    job_desc_extra = ['request_memory = 900']
    return HTCondorJobSubmitter(job_desc_extra=job_desc_extra)

@pytest.fixture()
def workingarea(tmpdir_factory):
    ret = mock.Mock(spec=WorkingArea)
    ret.path = str(tmpdir_factory.mktemp(''))
    ret.package_path.return_value = 'tpd_20161129_122841_HnpcmF'
    ret.executable = 'run.py'
    ret.extra_input_files = set(['python_modules.tar.gz', 'logging_levels.json.gz'])
    return ret

def test_repr(obj):
    repr(obj)

def test_run(obj, workingarea, proc_submit, caplog):
    with caplog.at_level(logging.WARNING):
        assert '1012.0' == obj.run(workingArea=workingarea, package_index=0)

    expected = textwrap.dedent("""
    executable = run.py
    output = results/$(resultdir)/stdout.$(cluster).$(process).txt
    error = results/$(resultdir)/stderr.$(cluster).$(process).txt
    log = results/$(resultdir)/log.$(cluster).$(process).txt
    arguments = $(resultdir).p.gz
    should_transfer_files = YES
    when_to_transfer_output = ON_EXIT
    transfer_input_files = $(resultdir).p.gz, logging_levels.json.gz, python_modules.tar.gz
    transfer_output_files = results
    universe = vanilla
    notification = Error
    getenv = True
    request_memory = 900
    queue resultdir in tpd_20161129_122841_HnpcmF
    """).strip()

    assert [mock.call(expected)] == proc_submit.communicate.call_args_list

##__________________________________________________________________||
def test_option_job_desc_dict(mocksubprocess, workingarea, proc_submit):
    job_desc_dict = collections.OrderedDict(
        [('request_memory', '1200'), ('Universe', 'chocolate')]
    )
    obj = HTCondorJobSubmitter(job_desc_dict=job_desc_dict)
    obj.run(workingArea=workingarea, package_index=0)

    expected = textwrap.dedent("""
    executable = run.py
    output = results/$(resultdir)/stdout.$(cluster).$(process).txt
    error = results/$(resultdir)/stderr.$(cluster).$(process).txt
    log = results/$(resultdir)/log.$(cluster).$(process).txt
    arguments = $(resultdir).p.gz
    should_transfer_files = YES
    when_to_transfer_output = ON_EXIT
    transfer_input_files = $(resultdir).p.gz, logging_levels.json.gz, python_modules.tar.gz
    transfer_output_files = results
    universe = chocolate
    notification = Error
    getenv = True
    request_memory = 1200
    queue resultdir in tpd_20161129_122841_HnpcmF
    """).strip()

    assert [mock.call(expected)] == proc_submit.communicate.call_args_list

##__________________________________________________________________||



import unittest

import io
import textwrap

##__________________________________________________________________||
hasPandas = False
try:
    import pandas as pd
    hasPandas = True
except ImportError:
    pass

##__________________________________________________________________||
tbl_01_txt = r"""
   v1        v2
   10       AAA
   20     "BBB"
   30     'BBB'
   30   \"BBB\"
   40   "\"BBB"
   50   "\"B BB"
   60    " xyz"
"""
tbl_01_txt = tbl_01_txt[tbl_01_txt.find('\n') + 1:] # remove 1st line
tbl_01_txt = textwrap.dedent(tbl_01_txt)
tbl_01_txt = tbl_01_txt.encode()

##__________________________________________________________________||
@unittest.skipUnless(hasPandas, "has no pandas")
class TestPandasQuotes(unittest.TestCase):

    def test_one(self):
        input = io.BytesIO(tbl_01_txt)
        tbl = pd.read_table(input, delim_whitespace = True, escapechar = '\\')

        expected = [
            'AAA',    # no quote
            'BBB',    # double quote (striped)
            "'BBB'",  # single quote (still there)
            '"BBB"',  # double quote (escaped)
            '"BBB',   # double quote (escaped) within double quote
            '"B BB',  # double quote (escaped with space)
            ' xyz'    # start with space
        ]

        actual = tbl.v2.values.tolist()
        self.assertEqual(expected, actual)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class Events(object):
    """An iterative object for events.

    You can use this class to iterate over entries in a ROOT TTree.

    You can instantiate this class with a TTree object and an
    optionally a maximum number of entries to loop over::

        inputFile = ROOT.TFile.Open(inputPath)
        tree = inputFile.Get(treeName)
        events = Events(tree)

    Then, the "for" loop for the tree entries can be::

        for event in events:

    Note: "event" and "events" are the same object. In each iteration,
    "event" (and "events") is loaded with the next entry in the tree.

    A content of the tree, e.g., a branch, can be accessed as an
    attribute of "event"::

          event.jet_pt

    In order to access to a particular entry, you can use an index.
    For example, to get 11th entry (the index for the first entry is
    0)::

        event = events[10]

    Note: Again "event" and "events" are the same object.

    """

    def __init__(self, tree, maxEvents=-1, start=0):

        if start < 0:
            raise ValueError("start must be greater than or equal to zero: {} is given".format(start))

        self.tree = tree
        nEventsInTree = self.tree.GetEntries()
        start = min(nEventsInTree, start)
        if maxEvents > -1:
            self.nEvents = min(nEventsInTree - start, maxEvents)
        else:
            self.nEvents = nEventsInTree - start
        self.maxEvents = maxEvents
        self.start = start
        self.iEvent = -1

    def __len__(self):
        return self.nEvents

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            self._repr_contents()
        )

    def _repr_contents(self):
        return 'tree = {!r}, maxEvents={!r}, start={!r}, nEvents={!r}, iEvent={!r}'.format(
            self.tree,
            self.maxEvents,
            self.start,
            self.nEvents,
            self.iEvent
        )

    def __getitem__(self, i):
        if i >= self.nEvents:
            self.iEvent = -1
            raise IndexError("the index is out of range: " + str(i))
        self.iEvent = i
        self.tree.GetEntry(self.start + self.iEvent)
        return self

    def __iter__(self):
        for self.iEvent in range(self.nEvents):
            self.tree.GetEntry(self.start + self.iEvent)
            yield self
        self.iEvent = -1

    def __getattr__(self, name):
        return getattr(self.tree, name)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock


from alphatwirl.configure import TableConfigCompleter

##__________________________________________________________________||
class MockDefaultSummary:
    pass

class MockSummary2:
    pass

class MockWeight:
    pass

class MockBinning:
    pass

##__________________________________________________________________||
defaultWeight = MockWeight()

@pytest.fixture()
def obj():
    return TableConfigCompleter(
        defaultSummaryClass=MockDefaultSummary,
        defaultWeight=defaultWeight,
        defaultOutDir='tmp'
    )

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

##__________________________________________________________________||
binning1 = MockBinning()
binning2 = MockBinning()

@pytest.mark.parametrize('arg, expected', [
    pytest.param(
        dict(),
        dict(
            keyAttrNames=(),
            keyIndices=None,
            binnings=None,
            keyOutColumnNames=(),
            valAttrNames=None,
            valIndices=None,
            summaryClass=MockDefaultSummary,
            valOutColumnNames=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_n.txt',
            outFilePath='tmp/tbl_n.txt',
        ),
        id='empty'
    ),
    pytest.param(
        dict(
            keyAttrNames=('met_pt', ),
            binnings=(binning1, ),
        ),
        dict(
            keyAttrNames=('met_pt',),
            keyIndices=None,
            binnings=(binning1, ),
            keyOutColumnNames=('met_pt',),
            valAttrNames=None,
            valIndices=None,
            summaryClass=MockDefaultSummary,
            valOutColumnNames=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_n.met_pt.txt',
            outFilePath='tmp/tbl_n.met_pt.txt',
        ),
        id='simple'
    ),
    pytest.param(
        dict(
            keyAttrNames=( ),
            binnings=( )
        ),
        dict(
            keyAttrNames=(),
            keyIndices=None,
            binnings=(),
            keyOutColumnNames=(),
            valAttrNames=None,
            valIndices=None,
            summaryClass=MockDefaultSummary,
            valOutColumnNames=('n', 'nvar'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_n.txt',
            outFilePath='tmp/tbl_n.txt',
        ),
        id='empty-key'
    ),
    pytest.param(
        dict(
            keyAttrNames=( ),
            binnings=( ),
            summaryClass=MockSummary2,
        ),
        dict(
            keyAttrNames=(),
            keyIndices=None,
            binnings=(),
            keyOutColumnNames=(),
            valAttrNames=None,
            valIndices=None,
            summaryClass=MockSummary2,
            valOutColumnNames=(),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_MockSummary2.txt',
            outFilePath='tmp/tbl_MockSummary2.txt',
        ),
        id='summary-class-empty-key-empty-val'
    ),
    pytest.param(
        dict(
            keyAttrNames=('key1', 'key2'),
            binnings=(binning1, binning2),
            summaryClass=MockSummary2,
        ),
        dict(
            keyAttrNames=('key1', 'key2'),
            keyIndices=None,
            binnings=(binning1, binning2),
            keyOutColumnNames=('key1', 'key2'),
            valAttrNames=None,
            valIndices=None,
            summaryClass=MockSummary2,
            valOutColumnNames=(),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_MockSummary2.key1.key2.txt',
            outFilePath='tmp/tbl_MockSummary2.key1.key2.txt',
        ),
        id='summary-class-2-keys-empty-vals'
    ),
    pytest.param(
        dict(
            keyAttrNames=('key1', 'key2'),
            binnings=(binning1, binning2),
            valAttrNames=('val1', 'val2'),
            summaryClass=MockSummary2,
        ),
        dict(
            keyAttrNames=('key1', 'key2'),
            keyIndices=None,
            binnings=(binning1, binning2),
            keyOutColumnNames=('key1', 'key2'),
            valAttrNames=('val1', 'val2'),
            valIndices=None,
            summaryClass=MockSummary2,
            valOutColumnNames=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_MockSummary2.key1.key2.val1.val2.txt',
            outFilePath='tmp/tbl_MockSummary2.key1.key2.val1.val2.txt',
        ),
        id='summary-class-2-keys-2-vals'
    ),
    pytest.param(
        dict(
            keyAttrNames=('key1', 'key2'),
            binnings=(binning1, binning2),
            keyIndices=(None, 1),
            valAttrNames=('val1', 'val2'),
            summaryClass=MockSummary2,
        ),
        dict(
            keyAttrNames=('key1', 'key2'),
            keyIndices=(None, 1),
            binnings=(binning1, binning2),
            keyOutColumnNames=('key1', 'key2'),
            valAttrNames=('val1', 'val2'),
            valIndices=None,
            summaryClass=MockSummary2,
            valOutColumnNames=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_MockSummary2.key1.key2-1.val1.val2.txt',
            outFilePath='tmp/tbl_MockSummary2.key1.key2-1.val1.val2.txt',
        ),
        id='summary-class-2-keys-2-vals-key-indices'
    ),
    pytest.param(
        dict(
            keyAttrNames=('key1', 'key2'),
            binnings=(binning1, binning2),
            valAttrNames=('val1', 'val2'),
            summaryClass=MockSummary2,
            valIndices=(2, None),
        ),
        dict(
            keyAttrNames=('key1', 'key2'),
            keyIndices=None,
            binnings=(binning1, binning2),
            keyOutColumnNames=('key1', 'key2'),
            valAttrNames=('val1', 'val2'),
            valIndices=(2, None),
            summaryClass=MockSummary2,
            valOutColumnNames=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_MockSummary2.key1.key2.val1-2.val2.txt',
            outFilePath='tmp/tbl_MockSummary2.key1.key2.val1-2.val2.txt',
        ),
        id='summary-class-2-keys-2-vals-val-indices'
    ),
    pytest.param(
        dict(
            keyAttrNames=('key1', 'key2'),
            binnings=(binning1, binning2),
            keyIndices=(None, 1),
            valAttrNames=('val1', 'val2'),
            summaryClass=MockSummary2,
            valIndices=(2, 3),
        ),
        dict(
            keyAttrNames=('key1', 'key2'),
            keyIndices=(None, 1),
            binnings=(binning1, binning2),
            keyOutColumnNames=('key1', 'key2'),
            valAttrNames=('val1', 'val2'),
            valIndices=(2, 3),
            summaryClass=MockSummary2,
            valOutColumnNames=('val1', 'val2'),
            weight=defaultWeight,
            sort=True,
            nevents=None,
            outFile=True,
            outFileName='tbl_MockSummary2.key1.key2-1.val1-2.val2-3.txt',
            outFilePath='tmp/tbl_MockSummary2.key1.key2-1.val1-2.val2-3.txt',
        ),
        id='summary-class-2-keys-2-vals-key-indices-val-indices'
    ),
])
def test_complete(obj, arg, expected):
    actual = obj.complete(arg)
    assert expected == actual
    assert arg is not actual

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import CommunicationChannel, TaskPackage

##__________________________________________________________________||
@pytest.fixture()
def dropbox():
    return mock.MagicMock()

@pytest.fixture()
def obj(dropbox):
    return CommunicationChannel(dropbox=dropbox)

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_begin_end(obj, dropbox):

    dropbox.open.assert_not_called()
    dropbox.close.assert_not_called()

    obj.begin()
    dropbox.open.assert_called_once()
    dropbox.close.assert_not_called()

    obj.begin()
    dropbox.open.assert_called_once() # don't open twice
    dropbox.close.assert_not_called()

    obj.end()
    dropbox.open.assert_called_once()
    dropbox.close.assert_called_once()

    obj.end()
    dropbox.open.assert_called_once()
    dropbox.close.assert_called_once() # don't close twice

    obj.begin()
    assert 2 == dropbox.open.call_count # can open again
    dropbox.close.assert_called_once()

def test_begin_terminate_end(obj, dropbox):
    obj.begin()
    assert 0 == dropbox.terminate.call_count
    obj.terminate()
    assert 1 == dropbox.terminate.call_count
    obj.end()

def test_put_when_closed(obj, dropbox, caplog):

    task1 = mock.MagicMock(name='task1')

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        obj.put(task1)

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'CommunicationChannel' in caplog.records[0].name
    assert 'the drop box is not open' in caplog.records[0].msg

    dropbox.put.assert_not_called()

def test_receive_when_closed(obj, dropbox, caplog):

    result1 = mock.MagicMock(name='result1')
    dropbox.receive = mock.MagicMock(return_value=[(0, result1)])

    with caplog.at_level(logging.WARNING, logger='alphatwirl'):
        result = obj.receive()

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert 'CommunicationChannel' in caplog.records[0].name
    assert 'the drop box is not open' in caplog.records[0].msg

    assert result is None

    obj.end()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.parallel import Parallel

##__________________________________________________________________||
@pytest.fixture()
def progressMonitor():
    return mock.Mock()

@pytest.fixture()
def communicationChannel():
    return mock.Mock()

@pytest.fixture()
def workingarea():
    return mock.Mock()

@pytest.fixture()
def obj(progressMonitor, communicationChannel, workingarea):
    return Parallel(
        progressMonitor = progressMonitor,
        communicationChannel = communicationChannel,
        workingarea = workingarea
    )

def test_repr(obj):
    repr(obj)

def test_workingarea(obj, workingarea):
    assert workingarea is obj.workingarea

def test_begin_terminate_end(obj, progressMonitor, communicationChannel):

    obj.begin()
    assert [mock.call()] == progressMonitor.begin.call_args_list
    assert [mock.call()] == communicationChannel.begin.call_args_list

    obj.terminate()
    assert [mock.call()] == communicationChannel.terminate.call_args_list

    obj.end()
    assert [mock.call()] == progressMonitor.end.call_args_list
    assert [mock.call()] == communicationChannel.end.call_args_list

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import itertools
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.selection.modules import AllwCount, AnywCount, NotwCount

##__________________________________________________________________||
class MockEventSelection(object):
    def begin(self, event): pass
    def __call__(self, event): pass
    def end(self): pass

##__________________________________________________________________||
def mk_tree():
    # all0 - all1 --- all2 --- sel1
    #              |        +- sel2
    #              +- not1 --- any1 --- all3 --- sel3
    #                                |        +- sel4
    #                                +- sel5

    sel1 = mock.Mock(spec=MockEventSelection)
    sel1.name ='sel1'
    sel2 = mock.Mock(spec=MockEventSelection)
    sel2.name ='sel2'
    sel3 = mock.Mock(spec=MockEventSelection)
    sel3.name ='sel3'
    sel4 = mock.Mock(spec=MockEventSelection)
    sel4.name ='sel4'
    sel5 = mock.Mock(spec=MockEventSelection)
    sel5.name ='sel5'

    all0 = AllwCount(name='all0')
    all1 = AllwCount(name='all1')
    all2 = AllwCount(name='all2')
    all3 = AllwCount(name='all3')
    any1 = AnywCount(name='any1')

    all3.add(sel3)
    all3.add(sel4)
    any1.add(all3)
    any1.add(sel5)

    not1 = NotwCount(any1, name='not1')

    all2.add(sel1)
    all2.add(sel2)

    all1.add(all2)
    all1.add(not1)

    all0.add(all1)

    return dict(
        alls=(all0, all1, all2, all3),
        anys=(any1, ),
        nots=(not1, ),
        sels=(sel1, sel2, sel3, sel4, sel5)
    )

##__________________________________________________________________||
def test_combination():

    tree = mk_tree()

    all0 = tree['alls'][0]
    sels = tree['sels']

    event = mock.Mock()
    all0.begin(event)

    all_possible_results = itertools.product(*[[True, False]]*len(sels))
    # e.g.,
    # [
    #     (True, True, True, True, True),
    #     (True, True, True, True, False),
    #     ...
    #     (False, False, False, False, False)
    # ]

    for l in all_possible_results:
        # e.g. l = (True, True, False, True, False)
        for sel, ret in zip(sels, l):
            sel.return_value = ret
        all0(event)

    all0.end()

    count = all0.results()
    assert [
        [1,          'AllwCount', 'all1',  3, 32],
        [2,          'AllwCount', 'all2',  8, 32],
        [3, 'MockEventSelection', 'sel1', 16, 32],
        [3, 'MockEventSelection', 'sel2',  8, 16],
        [2 ,         'NotwCount', 'not1',  3,  8],
        [3,          'AnywCount', 'any1',  5,  8],
        [4,          'AllwCount', 'all3',  2,  8],
        [5, 'MockEventSelection', 'sel3',  4,  8],
        [5, 'MockEventSelection', 'sel4',  2,  4],
        [4, 'MockEventSelection', 'sel5',  3,  6]
    ] == count._results

##__________________________________________________________________||
def test_merge():

    # manually call the fixture because multiple instances are needed

    # deep.copy() is not used because it will be difficult to access
    # to copied sels

    tree0 = mk_tree()
    all0 = tree0['alls'][0]

    tree0_copy1 = mk_tree()
    tree0_copy2 = mk_tree()

    all0_copy1 = tree0_copy1['alls'][0]
    all0_copy2 = tree0_copy2['alls'][0]

    sels_copy1 = tree0_copy1['sels']
    sels_copy2 = tree0_copy2['sels']

    event = mock.Mock()
    all0_copy1.begin(event)
    all0_copy2.begin(event)

    all_possible_results = list(itertools.product(*[[True, False]]*len(sels_copy1)))

    results1 = all_possible_results[:len(all_possible_results)//2]
    results2 = all_possible_results[len(all_possible_results)//2:]

    for l in results1:
        for sel, ret in zip(sels_copy1, l):
            sel.return_value = ret
        all0_copy1(event)

    for l in results2:
        for sel, ret in zip(sels_copy2, l):
            sel.return_value = ret
        all0_copy2(event)

    all0_copy1.end()
    all0_copy2.end()

    all0.merge(all0_copy1)
    all0.merge(all0_copy2)

    count = all0.results()
    assert [
        [1,          'AllwCount', 'all1',  3, 32],
        [2,          'AllwCount', 'all2',  8, 32],
        [3, 'MockEventSelection', 'sel1', 16, 32],
        [3, 'MockEventSelection', 'sel2',  8, 16],
        [2 ,         'NotwCount', 'not1',  3,  8],
        [3,          'AnywCount', 'any1',  5,  8],
        [4,          'AllwCount', 'all3',  2,  8],
        [5, 'MockEventSelection', 'sel3',  4,  8],
        [5, 'MockEventSelection', 'sel4',  2,  4],
        [4, 'MockEventSelection', 'sel5',  3,  6]
    ] == count._results

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events
    from alphatwirl.roottree import Branch

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def tree():
    input_file_name = 'sample_01.root'
    tree_name = 'tree'
    input_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', input_file_name)

    input_file = ROOT.TFile.Open(input_path)
    yield input_file.Get(tree_name)

@pytest.fixture()
def events(tree):
    yield Events(tree)

def test_branch(tree, events):
    jet_pt = events.jet_pt
    met_pt = events.met_pt
    assert isinstance(jet_pt, Branch)
    assert isinstance(met_pt, Branch)

    assert 0 == len(jet_pt)
    assert 1 == len(met_pt)
    assert 0.0 == met_pt[0]

    _assert_contents(tree, jet_pt, met_pt)
    _assert_contents(tree, jet_pt, met_pt) # assert twice

def test_2_events(tree, events):
    jet_pt = events.jet_pt
    met_pt = events.met_pt
    _assert_contents(tree, jet_pt, met_pt)

    # the 2nd events object from the same tree
    events2 = Events(tree)
    jet_pt2 = events2.jet_pt
    met_pt2 = events2.met_pt
    assert jet_pt is jet_pt2
    assert met_pt is met_pt2
    _assert_contents(tree, jet_pt, met_pt)

def _assert_contents(tree, jet_pt, met_pt):

    tree.GetEntry(0)
    assert 2 == len(jet_pt)
    assert 1 == len(met_pt)
    assert 124.55626678466797 == jet_pt[0]
    assert 86.90544128417969 == jet_pt[1]
    assert 43.783382415771484 == met_pt[0]

    tree.GetEntry(1)
    assert 3 == len(jet_pt)
    assert 1 == len(met_pt)
    assert 112.48554992675781 == jet_pt[0]
    assert 52.32780075073242 == jet_pt[1]
    assert 48.861289978027344 == jet_pt[2]
    assert 20.483951568603516 == met_pt[0]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

import numpy as np
import copy

##__________________________________________________________________||
class Sum(object):

    def __init__(self, val = None, weight = 1, contents = None):

        if contents is not None:
            self.contents = contents
            return

        if val is None:
            self.contents = [np.array([0])] # will be broadcasted when
                                            # added with more than 1
                                            # element
            return

        self.contents = [np.array(val)*weight]

    def __add__(self, other):
        contents = [self.contents[0] + other.contents[0]]
        return self.__class__(contents = contents)

    def __radd__(self, other):
        # is called with other = 0 when e.g. sum([obj1, obj2])
        if other == 0:
            return self.__class__() + self
        raise TypeError('unsupported: {!r} + {!r}'.format(other, self))

    def __repr__(self):
        return '{}(contents = {!r})'.format(self.__class__.__name__, self.contents)

    def __eq__(self, other):
        if len(self.contents) != len(other.contents): return False
        return all([np.all(self.contents[i] == other.contents[i]) for i in range(len(self.contents))])

    def __copy__(self):
        contents = [np.copy(self.contents[0])]
        return self.__class__(contents = contents)

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

from .parse_indices_config import parse_indices_config
from .BackrefMultipleArrayReader import BackrefMultipleArrayReader

##__________________________________________________________________||
class KeyValueComposer(object):
    """This class composes keys and values for the event

    (this docstring is under development.)

    This class can be used with BEvents.

    This class supports inclusive indices '*'

    This class supports back references.


    """
    def __init__(self, keyAttrNames=None, binnings=None, keyIndices=None,
                 valAttrNames=None, valIndices=None):

        # for __repr__()
        name_value_pairs = (
            ('keyAttrNames', keyAttrNames),
            ('binnings', binnings),
            ('keyIndices', keyIndices),
            ('valAttrNames', valAttrNames),
            ('valIndices', valIndices),
        )
        self._repr = '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

        key_attr_names = tuple(keyAttrNames) if keyAttrNames is not None else ()
        key_idxs = tuple(keyIndices) if keyIndices is not None else (None, )*len(key_attr_names)
        val_attr_names = tuple(valAttrNames) if valAttrNames is not None else ()
        val_idxs = tuple(valIndices) if valIndices is not None else (None, )*len(val_attr_names)

        if not len(key_attr_names) == len(key_idxs):
            raise ValueError(
                "the two tuples must have the same length: key_attr_names={}, key_idxs={}".format(
                    key_attr_names, key_idxs
                )
            )

        if not len(val_attr_names) == len(val_idxs):
            raise ValueError(
                "the two tuples must have the same length: val_attr_names={}, val_idxs={}".format(
                    val_attr_names, val_idxs
                )
            )

        self.binnings = tuple(binnings) if binnings is not None else None
        if self.binnings is not None and not len(key_attr_names) == len(self.binnings):
            raise ValueError(
                "the two tuples must have the same length: key_attr_names={}, self.binnings={}".format(
                    key_attr_names, self.binnings
                )
            )

        self._lenkey = len(key_attr_names)
        self.attr_names = key_attr_names + val_attr_names
        self.idxs_conf = key_idxs + val_idxs
        self.backref_idxs, self.idxs_conf = parse_indices_config(self.idxs_conf)

        self.ArrayReader = BackrefMultipleArrayReader

    def __repr__(self):
        return self._repr

    def begin(self, event):
        arrays = self._collect_arrays(event, self.attr_names)
        self.active = True if arrays is not None else False
        if not self.active: return
        self._array_reader = self.ArrayReader(arrays, self.idxs_conf, self.backref_idxs)

    def _collect_arrays(self, event, attr_names):
        ret = [ ]
        for varname in attr_names:
            try:
                attr = getattr(event, varname)
            except AttributeError as e:
                logger = logging.getLogger(__name__)
                logger.warning('{!r}: {!s}'.format(self, e))
                return None
            ret.append(attr)
        return ret

    def __call__(self, event):
        if not self.active: return ()

        try:
            arrays = self._array_reader.read()
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.error(e)
            logger.error(self)
            raise
        # e.g.,
        # arrays = (
        #     (1001, 15.3, -1.2, 20.2,  2.2, 0.1, 16.2, 22.1),
        #     (1001, 15.3, -1.2, 11.9,  1.2, 0.1, 16.2, 15.2),
        #     (1001, 15.3, -1.2, 13.3, -1.5, 0.1, 16.2, 16.3),
        #     (1001, 12.9,  5.2, 20.2,  2.2, 0.6, 13.1, 22.1),
        #     (1001, 12.9,  5.2, 11.9,  1.2, 0.6, 13.1, 15.2),
        #     (1001, 12.9,  5.2, 13.3, -1.5, 0.6, 13.1, 16.3),
        #     (1001,  9.2,  2.2, 20.2,  2.2, 1.2, 10.1, 22.1),
        #     (1001,  9.2,  2.2, 11.9,  1.2, 1.2, 10.1, 15.2),
        #     (1001,  9.2,  2.2, 13.3, -1.5, 1.2, 10.1, 16.3)
        # )


        # separate into keys and vals
        keyvals = tuple((e[:self._lenkey], e[self._lenkey:]) for e in arrays)
        # e.g.,
        # keyvals = (
        #     ((1001, 15.3, -1.2, 20.2,  2.2, 0.1), (16.2, 22.1)),
        #     ((1001, 15.3, -1.2, 11.9,  1.2, 0.1), (16.2, 15.2)),
        #     ((1001, 15.3, -1.2, 13.3, -1.5, 0.1), (16.2, 16.3)),
        #     ((1001, 12.9,  5.2, 20.2,  2.2, 0.6), (13.1, 22.1)),
        #     ((1001, 12.9,  5.2, 11.9,  1.2, 0.6), (13.1, 15.2)),
        #     ((1001, 12.9,  5.2, 13.3, -1.5, 0.6), (13.1, 16.3)),
        #     ((1001,  9.2,  2.2, 20.2,  2.2, 1.2), (10.1, 22.1)),
        #     ((1001,  9.2,  2.2, 11.9,  1.2, 1.2), (10.1, 15.2)),
        #     ((1001,  9.2,  2.2, 13.3, -1.5, 1.2), (10.1, 16.3))
        # )


        # apply binnings
        if self.binnings:
            keyvals = tuple((tuple(b(k) for b, k in zip(self.binnings, kk)), vv) for kk, vv in keyvals)
        # e.g.,
        # keyvals = (
        #     ((1001, 15,   -2, 20, None, 0.1), (16.2, 22.1)),
        #     ((1001, 15,   -2, 11,    1, 0.1), (16.2, 15.2)),
        #     ((1001, 15,   -2, 13,   -2, 0.1), (16.2, 16.3)),
        #     ((1001, 12, None, 20, None, 0.6), (13.1, 22.1)),
        #     ((1001, 12, None, 11,    1, 0.6), (13.1, 15.2)),
        #     ((1001, 12, None, 13,   -2, 0.6), (13.1, 16.3)),
        #     ((1001,  9,    2, 20, None, 1.2), (10.1, 22.1)),
        #     ((1001,  9,    2, 11,    1, 1.2), (10.1, 15.2)),
        #     ((1001,  9,    2, 13,   -2, 1.2), (10.1, 16.3))
        # )

        # remove None
        keyvals = tuple(e for e in keyvals if None not in e[0] and None not in e[1])
        # e.g.,
        # keyvals = (
        #     ((1001, 15, -2, 11,  1, 0.1), (16.2, 15.2)),
        #     ((1001, 15, -2, 13, -2, 0.1), (16.2, 16.3)),
        #     ((1001,  9,  2, 11,  1, 1.2), (10.1, 15.2)),
        #     ((1001,  9,  2, 13, -2, 1.2), (10.1, 16.3))
        # )

        return keyvals

##__________________________________________________________________||



import alphatwirl
import unittest

##__________________________________________________________________||
# How to run test
# at the directory three up (under AlphaTwirl/)
#
# to run all tests in this file
# $ python -m unittest tests.unit.examples.test_test_example
#
# to run a particular test in this file
# $ python -m unittest tests.unit.examples.test_test_example.TestExample1
#
# to run all tests in all files
# $ python -m unittest discover

##__________________________________________________________________||
class TestExample1(unittest.TestCase):
    def test_example(self):
        pass
    pass

##__________________________________________________________________||
class TestExample2(unittest.TestCase):
    def test_example(self):
        pass
    pass

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest
import time

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.progressbar import ProgressReporter

##__________________________________________________________________||
@pytest.fixture()
def queue():
    return mock.MagicMock()

@pytest.fixture()
def mocktime():
    return mock.MagicMock(return_value = 1000.0)

@pytest.fixture()
def reporter(queue, mocktime, monkeypatch):
    monkeypatch.setattr(time, 'time', mocktime)
    ret = ProgressReporter(queue)
    return ret

@pytest.fixture()
def report():
    return mock.MagicMock(**{'last.return_value': False, 'first.return_value': False})

@pytest.fixture()
def report_last():
    return mock.MagicMock(**{'last.return_value': True, 'first.return_value': False})

@pytest.fixture()
def report_first():
    return mock.MagicMock(**{'last.return_value': False, 'first.return_value': True})

##__________________________________________________________________||
def test_repr(reporter):
    repr(reporter)

##__________________________________________________________________||
def test_report_no_need_to_report(reporter, monkeypatch, report):

    mock_report = mock.MagicMock()
    monkeypatch.setattr(reporter, '_report', mock_report)

    mock_need_to_report = mock.MagicMock()
    mock_need_to_report.return_value = False
    monkeypatch.setattr(reporter, '_need_to_report', mock_need_to_report)

    reporter.report(report)

    mock_report.assert_not_called()

def test_report_need_to_report(reporter, monkeypatch, report):

    mock_report = mock.MagicMock()
    monkeypatch.setattr(reporter, '_report', mock_report)

    mock_need_to_report = mock.MagicMock()
    mock_need_to_report.return_value = True
    monkeypatch.setattr(reporter, '_need_to_report', mock_need_to_report)

    reporter.report(report)

    mock_report.assert_called_once_with(report)

##__________________________________________________________________||
def test__report(reporter, queue, mocktime, report):

    assert 1000.0 == reporter.last_time

    mocktime.return_value = 1000.2
    reporter._report(report)

    assert [mock.call(report)] == queue.put.call_args_list

    assert 1000.2 == reporter.last_time

def test_need_to_report(reporter, queue, mocktime, report, report_last,
                      report_first):

    interval = reporter.interval
    assert 0.1 == interval

    assert 1000.0 == reporter.last_time

    # before the interval passes
    mocktime.return_value += 0.1*interval
    assert not reporter._need_to_report(report)

    # the first report before the interval passes
    assert reporter._need_to_report(report_first)

    # the last report before the interval passes
    assert reporter._need_to_report(report_last)

    # after the interval passes
    mocktime.return_value += 1.2*interval
    assert reporter._need_to_report(report)

    assert 1000.0 == reporter.last_time

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import stat
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import SubprocessRunner

##__________________________________________________________________||
run_py_content = """
#!/usr/bin/env python
import time
import sys
import os
with open(os.path.join(sys.argv[1], 'sleep.txt'), 'r') as f:
    secs = f.read()
time.sleep(float(secs))
with open(os.path.join(sys.argv[1], 'result.txt'), 'w') as f:
    f.write('{} {} {}'.format(os.getpid(), sys.argv[1], secs))
"""
run_py_content = run_py_content.lstrip()

##__________________________________________________________________||
@pytest.fixture()
def taskdir(tmpdir_factory):
    ret = tmpdir_factory.mktemp('')
    ret = str(ret)
    path_run_py = os.path.join(ret, 'run.py')
    with open(path_run_py, 'w') as f:
        f.write(run_py_content)
        os.chmod(path_run_py, os.stat(path_run_py).st_mode | stat.S_IXUSR)
    return ret

def create_package(taskdir, package_path, sleep_time):
    os.makedirs(os.path.join(taskdir, package_path))
    with open(os.path.join(taskdir, package_path, 'sleep.txt'), 'w') as f:
            f.write(sleep_time)

@pytest.fixture()
def package0(taskdir):
    create_package(taskdir=taskdir, package_path='aaa', sleep_time='0.20')

@pytest.fixture()
def package1(taskdir):
    create_package(taskdir=taskdir, package_path='bbb', sleep_time='0.02')

@pytest.fixture()
def package2(taskdir):
    create_package(taskdir=taskdir, package_path='ccc', sleep_time='0.15')

@pytest.fixture()
def package3(taskdir):
    create_package(taskdir=taskdir, package_path='ddd', sleep_time='100.000')

@pytest.fixture()
def workingarea(taskdir, package0, package1, package2, package3):
    ret = mock.MagicMock(path=taskdir)
    package_path_dict = {0:'aaa', 1:'bbb', 2:'ccc', 3:'ddd'}
    ret.package_path.side_effect = lambda x: package_path_dict[x]
    return ret

def test_mockworkingarea(workingarea):
    assert 'aaa' == workingarea.package_path(0)
    assert 'bbb' == workingarea.package_path(1)
    assert 'ccc' == workingarea.package_path(2)

##__________________________________________________________________||
@pytest.fixture()
def obj():
    return SubprocessRunner()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_run_wait_terminate(obj, workingarea):

    pid1 = obj.run(workingarea, 0)
    pid2 = obj.run(workingarea, 1)
    pid3 = obj.run(workingarea, 2)

    assert {pid1, pid2, pid3} == set(obj.wait())
    # obj.wait() returns a list of finished pids, unsorted

    assert '{} aaa 0.20'.format(pid1) == open(os.path.join(workingarea.path, 'aaa', 'result.txt')).read()
    assert '{} bbb 0.02'.format(pid2) == open(os.path.join(workingarea.path, 'bbb', 'result.txt')).read()
    assert '{} ccc 0.15'.format(pid3) == open(os.path.join(workingarea.path, 'ccc', 'result.txt')).read()

    obj.failed_runids([])

    obj.terminate()

def test_run_poll_terminate(obj, workingarea):
    pid1 = obj.run(workingarea, 0)
    pid2 = obj.run(workingarea, 1)
    pid3 = obj.run(workingarea, 2)
    obj.poll()
    obj.terminate()

def test_run_terminate(obj, workingarea):
    pid1 = obj.run(workingarea, 0)
    pid2 = obj.run(workingarea, 1)
    pid3 = obj.run(workingarea, 2)
    obj.terminate()

def test_wait_terminate(obj):
    assert [ ] == obj.wait()
    obj.terminate()

def test_poll_terminate_long_task(obj, workingarea):
    pid = obj.run(workingarea, 3)
    obj.poll()
    obj.terminate()
    result_path = os.path.join(workingarea.path, 'ddd', 'result.txt')
    assert not os.path.isfile(result_path)

def test_terminate_long_task(obj, workingarea):
    pid = obj.run(workingarea, 3)
    obj.terminate()
    result_path = os.path.join(workingarea.path, 'ddd', 'result.txt')
    assert not os.path.isfile(result_path)

##__________________________________________________________________||
def test_run_multiple_wait_terminate(obj, workingarea):

    pids = obj.run_multiple(workingarea, [0, 1, 2])

    assert set(pids) == set(obj.wait())
    # obj.wait() returns a list of finished pids, unsorted

    assert '{} aaa 0.20'.format(pids[0]) == open(os.path.join(workingarea.path, 'aaa', 'result.txt')).read()
    assert '{} bbb 0.02'.format(pids[1]) == open(os.path.join(workingarea.path, 'bbb', 'result.txt')).read()
    assert '{} ccc 0.15'.format(pids[2]) == open(os.path.join(workingarea.path, 'ccc', 'result.txt')).read()

    obj.failed_runids([])

    obj.terminate()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import logging
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True


if not has_no_ROOT:
    from alphatwirl.roottree import EventBuilderConfig
    from alphatwirl.roottree import EventBuilder

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def mockroot(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['alphatwirl.roottree.EventBuilder']
    monkeypatch.setattr(module, 'ROOT', ret)
    return ret

@pytest.fixture()
def mockevents(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['alphatwirl.roottree.BEventBuilder']
    monkeypatch.setattr(module, 'BEvents', ret)
    module = sys.modules['alphatwirl.roottree.EventBuilder']
    monkeypatch.setattr(module, 'Events', ret)
    return ret

@pytest.fixture()
def mocktfile():
    ret = mock.Mock()
    ret.IsZombie.return_value = False
    return ret

@pytest.fixture()
def mocktfile_null():
    ret = mock.Mock()
    ret.GetName.side_effect = ReferenceError
    return ret

@pytest.fixture()
def mocktfile_zombie():
    ret = mock.Mock()
    ret.IsZombie.return_value = True
    return ret

@pytest.fixture()
def obj(request, mockroot, mockevents):
    config = EventBuilderConfig(
        inputPaths=['/path/to/input1/tree.root', '/path/to/input2/tree.root'],
        treeName='tree',
        maxEvents=123,
        start=11,
        name='TTJets'
    )
    return EventBuilder(config, EventsClass=mockevents)

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_build(obj, mockroot, mocktfile, mockevents):
    mockroot.TFile.Open.return_value = mocktfile
    events = obj()
    assert [mock.call('tree')] == mockroot.TChain.call_args_list
    chain = mockroot.TChain()
    assert [
        mock.call('/path/to/input1/tree.root'),
        mock.call('/path/to/input2/tree.root'),
    ] == chain.Add.call_args_list
    assert [mock.call(chain, 123, 11)] == mockevents.call_args_list

def test_build_raise_null_file(obj, mockroot, mocktfile_null, caplog):
    mockroot.TFile.Open.return_value = mocktfile_null
    with caplog.at_level(logging.WARNING):
        with pytest.raises(OSError):
            events = obj()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    ## assert 'BEventBuilder' in caplog.records[0].name
    assert 'cannot open' in caplog.records[0].msg

def test_build_raise_zombie_file(obj, mockroot, mocktfile_zombie, caplog):
    mockroot.TFile.Open.return_value = mocktfile_zombie
    with caplog.at_level(logging.WARNING):
        with pytest.raises(OSError):
            events = obj()
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    ## assert 'BEventBuilder' in caplog.records[0].name
    assert 'cannot open' in caplog.records[0].msg

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import logging

import ROOT

##__________________________________________________________________||
def is_ROOT_null_pointer(tobject):
    try:
        tobject.GetName()
        return False
    except ReferenceError:
        return True

##__________________________________________________________________||
def get_entries_in_tree_in_file(path, tree_name, raises=False):

    ##
    file_ = ROOT.TFile.Open(path)
    if is_ROOT_null_pointer(file_) or file_.IsZombie():
        logger = logging.getLogger(__name__)
        msg = 'cannot open {}'.format(path)
        if raises:
            logger.error(msg)
            raise OSError(msg)
        logger.warning(msg)
        return None

    ##
    tree = file_.Get(tree_name)
    if is_ROOT_null_pointer(tree):
        logger = logging.getLogger(__name__)
        msg = 'cannot find tree "{}" in {}'.format(tree_name, path)
        if raises:
            logger.error(msg)
            raise ReferenceError(msg)
        logger.warning(msg)
        return None

    ##
    ret = tree.GetEntriesFast()
    file_.Close()
    return ret

##__________________________________________________________________||
def inspect_tree(tree):
    ret = { }
    ret['leaves'] = [inspect_leaf(leaf) for leaf in tree.GetListOfLeaves()]
    return ret

##__________________________________________________________________||
def inspect_leaf(leaf):
    ret = { }
    ret.update(inspect_leaf_definition(leaf))
    ret.update(inspect_leaf_size(leaf))
    return ret

##__________________________________________________________________||
def inspect_leaf_definition(leaf):
    leafcount = leaf.GetLeafCount()
    isArray = not is_ROOT_null_pointer(leafcount)
    ret = { }
    ret['name'] = leaf.GetName()
    ret['type'] = leaf.GetTypeName()
    ret['isarray'] = '1' if isArray else '0'
    ret['countname'] = leafcount.GetName() if isArray else None
    ret['title'] = leaf.GetBranch().GetTitle()
    return ret

##__________________________________________________________________||
def inspect_leaf_size(leaf):
    ret = { }
    zipbytes = leaf.GetBranch().GetZipBytes()/1024.0/1024.0 # MB
    totalsize = leaf.GetBranch().GetTotalSize()/1024.0/1024.0 # MB
    ret['size'] = zipbytes
    ret['uncompressed_size'] = totalsize
    ret['compression_factor'] = totalsize/zipbytes if zipbytes > 0 else 0
    return ret

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
import logging
import pytest
from collections import deque

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.concurrently import TaskPackageDropbox

##__________________________________________________________________||
@pytest.fixture()
def packages():
    return [
        mock.Mock(name='package0'),
        mock.Mock(name='package1'),
        mock.Mock(name='package2'),
        mock.Mock(name='package3'),
        mock.Mock(name='package4'),
    ]

@pytest.fixture()
def pkgidx_result_pairs(results):
    # e.g., [(0, package0), (1, package1), ...]
    return list(enumerate(results))

@pytest.fixture()
def results():
    return [
        mock.Mock(name='result0'),
        mock.Mock(name='result1'),
        mock.Mock(name='result2'),
        mock.Mock(name='result3'),
        mock.Mock(name='result4'),
    ]

@pytest.fixture()
def collect_results(results):
    ret = {i: deque([r]) for i, r in enumerate(results)}
    # e.g., {0: deque([result0]), 1:deque([result1]), ...}

    # 3rd job fails twice before succeed
    ret[2].extendleft([None, None])
    # e.g., deque([None, None, result2])

    return ret

@pytest.fixture()
def workingarea(collect_results):
    ret = mock.MagicMock()
    ret.put_package.side_effect = [0, 1, 2, 3, 4] # package indices
    ret.collect_result.side_effect = lambda i: collect_results[i].popleft()
    return ret

@pytest.fixture()
def dispatcher():
    ret = mock.MagicMock()
    ret.run_multiple.return_value = [1000, 1001, 1002, 1003, 1004]

    # jobs finish in steps
    ret.poll.side_effect = [
        [1001, 1003],
        [ ],
        [1002],
        [1000, 1005],
        [1006, 1004]
    ]

    # for resubmission
    ret.run.side_effect = [1005, 1006]

    return ret

@pytest.fixture()
def mock_sleep(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(time, 'sleep', ret)
    return ret

@pytest.fixture()
def obj(workingarea, dispatcher, packages, mock_sleep):
    ret = TaskPackageDropbox(workingArea=workingarea, dispatcher=dispatcher, sleep=0.01)
    ret.open()
    ret.put_multiple(packages)
    yield ret
    ret.close()

##__________________________________________________________________||
def test_receive(obj, mock_sleep, pkgidx_result_pairs):
    assert pkgidx_result_pairs == obj.receive()
    assert [mock.call(0.01)]*4 == mock_sleep.call_args_list

def test_receive_dispatcher_received_failed_runids(obj, mock_sleep, dispatcher):
    obj.receive()
    assert [mock.call(0.01)]*4 == mock_sleep.call_args_list
    assert [mock.call([1002]), mock.call([1005])] == dispatcher.failed_runids.call_args_list

def test_receive_logging_resubmission(obj, mock_sleep, caplog):
    with caplog.at_level(logging.WARNING):
        obj.receive()
    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[1].levelname == 'WARNING'
    assert 'TaskPackageDropbox' in caplog.records[0].name
    assert 'TaskPackageDropbox' in caplog.records[1].name
    assert 'resubmitting' in caplog.records[0].msg
    assert 'resubmitting' in caplog.records[1].msg
    assert [mock.call(0.01)]*4 == mock_sleep.call_args_list

def test_receive_in_one_step(obj, mock_sleep, pkgidx_result_pairs, dispatcher, collect_results):
    # make all jobs finish by the first poll
    dispatcher.poll.side_effect = [[1000, 1001, 1002, 1003, 1004]]

    # make the 3rd job successful by removing two None's
    collect_results[2].popleft() # deque([None, result2])
    collect_results[2].popleft() # deque([result2])

    assert pkgidx_result_pairs == obj.receive()
    assert [ ] == mock_sleep.call_args_list

##__________________________________________________________________||
def test_poll(obj, mock_sleep, pkgidx_result_pairs):
    actual = [ ]
    while len(actual) < len(pkgidx_result_pairs):
        actual.extend(obj.poll())
    assert sorted(pkgidx_result_pairs) == sorted(actual)
    assert [ ] == mock_sleep.call_args_list

def test_poll_then_receive(obj, mock_sleep, pkgidx_result_pairs):
    actual = [ ]
    actual.extend(obj.poll())
    actual.extend(obj.receive())

    assert sorted(pkgidx_result_pairs) == sorted(actual)
    assert [mock.call(0.01)]*3 == mock_sleep.call_args_list

##__________________________________________________________________||
def test_receive_one(obj, mock_sleep, pkgidx_result_pairs):
    actual = [ ]
    while len(actual) < len(pkgidx_result_pairs):
        actual.append(obj.receive_one())
    assert obj.receive_one() is None
    assert sorted(pkgidx_result_pairs) == sorted(actual)
    assert [mock.call(0.01)]*4 == mock_sleep.call_args_list

def test_receive_one_single_call(obj, mock_sleep, pkgidx_result_pairs, workingarea):
    pair = obj.receive_one()
    assert pair in pkgidx_result_pairs
    assert [mock.call(pair[0])] == workingarea.collect_result.call_args_list
    assert [] == mock_sleep.call_args_list

@pytest.mark.parametrize('dispatcher_poll', [
    pytest.param(
        [[1001, 1003], [ ], [1002], [1000, 1005], [1006, 1004]],
        id='as_in_fixture'
    ),
    pytest.param(
        [[1001, 1003], [ ], [1002], [1000, 1005], [1006], [1004]],
        id='last_one'
    ),
    pytest.param(
        [[ ], [1001, 1003], [ ], [1002], [1000, 1005], [1006, 1004]],
        id='empty_first'
    ),
])
def test_receive_one_param(obj, mock_sleep, pkgidx_result_pairs, dispatcher, dispatcher_poll):
    dispatcher.poll.side_effect = dispatcher_poll

    actual = [ ]
    while len(actual) < len(pkgidx_result_pairs):
        actual.append(obj.receive_one())
    assert obj.receive_one() is None
    assert sorted(pkgidx_result_pairs) == sorted(actual)
    assert [mock.call(0.01)]*(len(dispatcher_poll)-1) == mock_sleep.call_args_list

def test_receive_one_then_receive(obj, mock_sleep, pkgidx_result_pairs):
    actual = [ ]

    actual.append(obj.receive_one())

    actual.extend(obj.receive())

    assert sorted(pkgidx_result_pairs) == sorted(actual)
    assert [mock.call(0.01)]*4 == mock_sleep.call_args_list

def test_receive_one_then_poll(obj, mock_sleep, pkgidx_result_pairs):
    actual = [ ]

    actual.append(obj.receive_one())
    actual.append(obj.receive_one())

    actual.extend(obj.poll())
    actual.extend(obj.poll())
    actual.extend(obj.poll())
    actual.extend(obj.poll())

    assert sorted(pkgidx_result_pairs) == sorted(actual)
    assert [mock.call(0.01)] == mock_sleep.call_args_list

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

from alphatwirl.roottree import Events

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as BEvents


##__________________________________________________________________||
events_classes = [Events]
if not has_no_ROOT:
    events_classes.append(BEvents)
events_classes_ids = [c.__name__ for c in events_classes]

##__________________________________________________________________||
class MockFile(object):
    pass

class MockTree(object):
    def __init__(self, entries=100):
        self.entries = entries
        self.ievent = -1
        self.branchstatus = [ ]
        self.branch1 = 1111
        self.directory = MockFile()
    def GetDirectory(self):
        return self.directory
    def GetEntries(self):
        return self.entries
    def GetEntry(self, ientry):
        if ientry < self.entries:
            nbytes = 10
            self.ievent = ientry
        else:
            nbytes = 0
            self.ievent = -1
        return nbytes

    def SetBranchStatus(self, bname, status):
        self.branchstatus.append((bname, status))

def test_mocktree():

    tree = MockTree(entries=3)
    assert isinstance(tree.GetDirectory(), MockFile)
    assert 3 == tree.GetEntries()

    assert -1 == tree.ievent

    nbytes = 10
    assert nbytes == tree.GetEntry(0)
    assert 0 == tree.ievent
    assert nbytes == tree.GetEntry(1)
    assert 1 == tree.ievent
    assert nbytes == tree.GetEntry(2)
    assert 2 == tree.ievent
    assert 0 == tree.GetEntry(3)
    assert -1 == tree.ievent

##__________________________________________________________________||
@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_init(Events):
    tree = MockTree()
    events = Events(tree)
    events = Events(tree, 100)
    assert tree is events.tree

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_repr(Events):
    tree = MockTree()
    events = Events(tree)
    repr(events)

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_nEvents_default(Events):
    tree = MockTree(entries=100)
    events = Events(tree)
    assert 100 == events.nEvents # default the same as entries
    assert 100 == len(events)

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
@pytest.mark.parametrize('maxEvents, expected ', [
    pytest.param(-1, 100, id='default'),
    pytest.param(50, 50, id='less'),
    pytest.param(120, 100, id='more'),
    pytest.param(100, 100, id='exact'),
])
def test_nEvents(Events, maxEvents, expected):
    tree = MockTree(entries=100)
    events = Events(tree, maxEvents)
    assert expected == events.nEvents
    assert expected == len(events)

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
@pytest.mark.parametrize('maxEvents, start, expected ', [
    pytest.param(-1, 1, 99, id='all_events_start_2nd'),
    pytest.param(10, 1, 10, id='nEvents_equal_maxEvents'),
    pytest.param(-1, 99, 1, id='all_events_start_last'),
    pytest.param(20, 99, 1, id='nEvents_less_than_maxEvents'),
    pytest.param(-1, 100, 0, id='nEvents_zero_1'),
    pytest.param(-1, 110, 0, id='nEvents_zero_2'),
    pytest.param(10, 100, 0, id='nEvents_zero_3'),
])
def test_nEvents_start(Events, maxEvents, start, expected):
    tree = MockTree(entries=100)
    events = Events(tree, maxEvents=maxEvents, start=start)
    assert expected == events.nEvents
    assert expected == len(events)

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_nEvents_start_raise(Events):
    tree = MockTree(entries=100)
    with pytest.raises(ValueError):
        Events(tree, maxEvents=-1, start=-10)

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_iter_iEvent(Events):

    tree = MockTree(entries=4)
    events = Events(tree)
    assert -1 == events.iEvent

    it = iter(events)
    event = next(it)
    assert 0 == event.iEvent
    assert 0 == tree.ievent
    event = next(it)
    assert 1 == event.iEvent
    assert 1 == tree.ievent
    event = next(it)
    assert 2 == event.iEvent
    assert 2 == tree.ievent
    event = next(it)
    assert 3 == event.iEvent
    assert 3 == tree.ievent
    with pytest.raises(StopIteration):
        next(it)
    assert -1 == event.iEvent

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_iter_maxEvents(Events):
    tree = MockTree(entries=40)
    events = Events(tree, maxEvents=4)
    assert -1 == events.iEvent

    it = iter(events)
    event = next(it)
    assert 0 == event.iEvent
    event = next(it)
    assert 1 == event.iEvent
    event = next(it)
    assert 2 == event.iEvent
    event = next(it)
    assert 3 == event.iEvent
    with pytest.raises(StopIteration):
        next(it)
    assert -1 == event.iEvent

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_iter_iEvent_start(Events):
    tree = MockTree(entries=4)
    events = Events(tree, start=2)
    assert -1 == events.iEvent

    it = iter(events)
    event = next(it)
    assert 0 == event.iEvent
    assert 2 == tree.ievent
    event = next(it)
    assert 1 == event.iEvent
    assert 3 == tree.ievent
    with pytest.raises(StopIteration):
        next(it)
    assert -1 ==event.iEvent

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_iter_maxEvents_start(Events):

    tree = MockTree(entries=40)
    events = Events(tree, maxEvents=4, start=2)
    assert -1 == events.iEvent

    it = iter(events)
    event = next(it)
    assert 0 == event.iEvent
    assert 2 == tree.ievent
    event = next(it)
    assert 1 == event.iEvent
    assert 3 == tree.ievent
    event = next(it)
    assert 2 == event.iEvent
    assert 4 == tree.ievent
    event = next(it)
    assert 3 == event.iEvent
    assert 5 == tree.ievent
    with pytest.raises(StopIteration):
        next(it)
    assert -1 == event.iEvent

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_getitem(Events):
    tree = MockTree(entries=4)
    events = Events(tree)
    assert -1 == events.iEvent

    event = events[0]
    assert 0 == event.iEvent
    assert 0 == tree.ievent
    event = events[1]
    assert 1 == event.iEvent
    assert 1 == tree.ievent
    event = events[2]
    assert 2 == event.iEvent
    assert 2 == tree.ievent
    event = events[3]
    assert 3 == event.iEvent
    assert 3 == tree.ievent
    with pytest.raises(IndexError):
        events[4]
    assert -1 == events.iEvent

@pytest.mark.parametrize('Events', events_classes, ids=events_classes_ids)
def test_getitem_start(Events):
    tree = MockTree(entries=4)
    events = Events(tree, start=2)
    assert -1 == events.iEvent

    event = events[0]
    assert 0 == event.iEvent
    assert 2 == tree.ievent
    event = events[1]
    assert 1 == event.iEvent
    assert 3 == tree.ievent
    with pytest.raises(IndexError):
        events[4]
    assert -1 == events.iEvent

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from alphatwirl.loop import EventLoop
from alphatwirl import progressbar

##__________________________________________________________________||
@pytest.fixture(params=[0, 3])
def events(request):
    nevents = request.param
    ret = [mock.Mock(name='event{}'.format(i)) for i in range(nevents)]
    # e.g., [mock.Mock(name='event0'), mock.Mock(name='event1'), ...]
    return ret

@pytest.fixture()
def build_events(events):
    ret = mock.Mock()
    ret.return_value = events
    return ret

@pytest.fixture()
def reader():
    return mock.Mock()

@pytest.fixture()
def obj(build_events, reader):
    return EventLoop(build_events, reader)

##__________________________________________________________________||
def test_progressbar_label(build_events, reader):
    obj = EventLoop(build_events, reader)
    assert 'EventLoop' == obj.progressbar_label

    obj = EventLoop(build_events, reader, progressbar_label='TTJets')
    assert 'TTJets' == obj.progressbar_label

def test_repr(obj):
    repr(obj)

def test_call(obj, events, reader):

    assert reader == obj()

    expected = [ ]
    expected.append(mock.call.begin(events)) # begin will called even
                                             # if len(events) == 0
    expected.extend([mock.call.event(e) for e in events])
    expected.append(mock.call.end())
    assert expected == reader.method_calls

##__________________________________________________________________||
@pytest.fixture()
def report_progress(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['alphatwirl.progressbar']
    monkeypatch.setattr(module, 'report_progress', ret)
    return ret

@pytest.fixture()
def ProgressReport(monkeypatch):
    ret = mock.Mock()
    module = sys.modules['alphatwirl.progressbar']
    monkeypatch.setattr(module, 'ProgressReport', ret)
    return ret

def test_report_progress(obj, report_progress, ProgressReport):
    obj()
    expected = [
        mock.call(ProgressReport(taskid=obj.taskid, name='EventLoop', done=0, total=3)),
        mock.call(ProgressReport(taskid=obj.taskid, name='EventLoop', done=1, total=3)),
        mock.call(ProgressReport(taskid=obj.taskid, name='EventLoop', done=2, total=3)),
        mock.call(ProgressReport(taskid=obj.taskid, name='EventLoop', done=3, total=3))
    ]
    actual = report_progress.call_args_list
    assert expected[0] == actual[0]

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os
import sys
import pytest

has_no_ROOT = False
try:
    import ROOT
except ImportError:
    has_no_ROOT = True

if not has_no_ROOT:
    from alphatwirl.roottree import BEvents as Events
    from alphatwirl.roottree import Branch

##__________________________________________________________________||
pytestmark = pytest.mark.skipif(has_no_ROOT, reason="has no ROOT")

##__________________________________________________________________||
@pytest.fixture()
def chain():
    tree_name = 'tree'
    chain = ROOT.TChain(tree_name)
    # add no files to the chain
    yield chain

@pytest.fixture()
def events(chain):
    yield Events(chain)

def test_access_branch_raise(events):
    with pytest.raises(AttributeError):
        events.var

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class NextKeyComposer(object):
    def __init__(self, binnings):
        self._binnings = binnings

    def __repr__(self):
        return '{}(binnings = {!r})'.format(
            self.__class__.__name__,
            self._binnings
        )

    def __call__(self, key):
        """returns a list of the next keys

        e.g.,
        If key = (11, 8, 20)
        it returns ((12, 8, 20), (11, 9, 20), (11, 8, 21))

        When the next bin is None, it won't be included.

        When the next bin is the same, it won't be included.

        """
        ret = [ ]
        for i in range(len(self._binnings)):
            keyc = list(key)
            thisbin = keyc[i]
            nextbin = self._binnings[i].next(thisbin)
            if nextbin is None: continue
            if nextbin == thisbin: continue
            keyc[i] = nextbin
            ret.append(tuple(keyc))
        return tuple(ret)

##__________________________________________________________________||



from .ToTupleList import ToTupleList
from .ToTupleListWithDatasetColumn import ToTupleListWithDatasetColumn
from .WriteListToFile import WriteListToFile

##__________________________________________________________________||
hasPandas = False
try:
    import pandas
    hasPandas = True
except ImportError:
    pass

if hasPandas:
    from .ToDataFrameWithDatasetColumn import ToDataFrameWithDatasetColumn
    from .ToDataFrame import ToDataFrame
    from .WritePandasDataFrameToFile import WritePandasDataFrameToFile

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import copy
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

import alphatwirl
from alphatwirl.summary import Reader

##__________________________________________________________________||
@pytest.fixture()
def mockKeyValComposer():
    ## ret = mock.Mock(spec=alphatwirl.summary.KeyValueComposer)
    ## ideally, prefer to use spec. However, mock.Mock with spec
    ## sometimes cannot be copied or deep copied in python 2 with mock
    ## 2.0.0. see tests/unit/examples/test_mock.py
    ret = mock.Mock()
    return ret

@pytest.fixture()
def mockSummarizer():
    return mock.MagicMock()

@pytest.fixture()
def mockCollector():
    return mock.Mock()

@pytest.fixture()
def mockNextKeyComposer():
    return mock.Mock()

@pytest.fixture()
def mockWeightCalculator():
    return mock.Mock()

@pytest.fixture()
def obj(mockKeyValComposer, mockSummarizer, mockCollector, mockNextKeyComposer, mockWeightCalculator):
    return Reader(
        mockKeyValComposer, mockSummarizer,
        nextKeyComposer=mockNextKeyComposer,
        collector=mockCollector,
        weightCalculator=mockWeightCalculator
    )

def test_init_without_collector(mockKeyValComposer, mockSummarizer, mockNextKeyComposer, mockWeightCalculator):
    ## possible to init without collector for backward compatibility
    ##
    obj = Reader(
        mockKeyValComposer, mockSummarizer,
        nextKeyComposer=mockNextKeyComposer,
        weightCalculator=mockWeightCalculator
    )

def test_repr(obj):
    repr(obj)

def test_begin(obj, mockKeyValComposer):
    event = mock.Mock()
    obj.begin(event)
    assert [mock.call(event)] == mockKeyValComposer.begin.call_args_list

def test_event_raise(obj, mockKeyValComposer):
    event = mock.Mock()
    mockKeyValComposer.side_effect = Exception
    with pytest.raises(Exception):
        obj.event(event)

def test_event(obj, mockKeyValComposer, mockSummarizer, mockWeightCalculator):
    event = mock.Mock()
    key1 = mock.Mock(name='key1')
    val1 = mock.Mock(name='val1')
    key2 = mock.Mock(name='key2')
    val2 = mock.Mock(name='val2')
    mockKeyValComposer.return_value = [(key1, val1), (key2, val2)]
    weight1 = mock.Mock(name='weight1')
    mockWeightCalculator.return_value = weight1
    obj.event(event)

    assert [
        mock.call(key=key1, val=val1, weight=weight1),
        mock.call(key=key2, val=val2, weight=weight1),
    ] == mockSummarizer.add.call_args_list

    assert [mock.call(event)] == mockKeyValComposer.call_args_list
    assert [mock.call(event)] == mockWeightCalculator.call_args_list

def test_event_nevents(mockKeyValComposer, mockSummarizer, mockWeightCalculator):
    obj = Reader(
        mockKeyValComposer, mockSummarizer,
        weightCalculator=mockWeightCalculator,
        nevents=2 # read only first 2 events
    )

    key1 = mock.Mock(name='key1')
    val1 = mock.Mock(name='val1')
    key2 = mock.Mock(name='key2')
    val2 = mock.Mock(name='val2')
    key3 = mock.Mock(name='key3')
    val3 = mock.Mock(name='val3')
    mockKeyValComposer.side_effect = [[(key1, val1)], [(key2, val2)], [(key3, val3)]]
    weight1 = mock.Mock(name='weight1')
    mockWeightCalculator.return_value = weight1

    event1 = mock.Mock(name='event1')
    event2 = mock.Mock(name='event2')
    event3 = mock.Mock(name='event3')
    obj.event(event1)
    obj.event(event2)
    obj.event(event3)

    assert [
        mock.call(key=key1, val=val1, weight=weight1),
        mock.call(key=key2, val=val2, weight=weight1),
    ] == mockSummarizer.add.call_args_list

    assert [mock.call(event1), mock.call(event2)] == mockKeyValComposer.call_args_list
    assert [mock.call(event1), mock.call(event2)] == mockWeightCalculator.call_args_list

def test_end(obj, mockSummarizer, mockNextKeyComposer):
    key1 = mock.MagicMock(name='key1')
    key2 = mock.MagicMock(name='key2')
    key3 = mock.MagicMock(name='key3')

    key1.__lt__.return_value = True
    key2.__lt__.return_value = True
    key3.__lt__.return_value = True

    key11 = mock.Mock(name='key12')
    key21 = mock.Mock(name='key21')
    key22 = mock.Mock(name='key22')

    mockSummarizer.keys.return_value = [key1, key2, key3]
    mockNextKeyComposer.side_effect = [(key11, ), (key21, key22), ()]
    obj.end()
    assert [
        mock.call(key11), mock.call(key21), mock.call(key22)
    ] == mockSummarizer.add_key.call_args_list

def test_end_None_nextKeyComposer(mockKeyValComposer, mockSummarizer):
    obj = Reader(
        mockKeyValComposer, mockSummarizer,
        nextKeyComposer=None
    )
    key1 = mock.Mock(name='key1')
    key2 = mock.Mock(name='key2')
    key3 = mock.Mock(name='key3')
    mockSummarizer.keys.return_value = [key1, key2, key3]
    obj.end()
    assert [ ] == mockSummarizer.add_key.call_args_list

def test_results(obj, mockSummarizer):
    assert mockSummarizer is obj.results()

def test_merge(obj, mockSummarizer):
    obj1 = copy.deepcopy(obj)

    assert obj1.summarizer is not obj.summarizer
    # note: if just print summarizers like
    # print obj1.summarizer, obj.summarizer
    # the same ids are shown, e.g.,
    # <MagicMock id='4782590736'> <MagicMock id='4782590736'>
    # this is very confusing. they are indeed different objects.
    # the attribute `id` is copied.
    # to print the real id of the copied object, need to print repr()
    # of it.
    # print repr(obj1.summarizer)
    # <MagicMock id='4782590352'>

    obj.merge(obj1)

    assert [mock.call(obj1.summarizer)] == mockSummarizer.__iadd__.call_args_list
    assert obj.summarizer is mockSummarizer.__iadd__()

def test_collect(obj, mockCollector):
    result = obj.collect()
    assert [mock.call(obj)] == mockCollector.call_args_list
    assert mockCollector() is result

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
def expand_path_cfg(path_cfg, alias_dict={ }, overriding_kargs={ }):
    """expand a path config

    Args:
        path_cfg (str, tuple, dict): a config for path
        alias_dict (dict): a dict for aliases
        overriding_kargs (dict): to be used for recursive call
    """

    if isinstance(path_cfg, str):
        return _expand_str(path_cfg, alias_dict, overriding_kargs)

    if isinstance(path_cfg, dict):
        return _expand_dict(path_cfg, alias_dict)

    # assume tuple or list
    return _expand_tuple(path_cfg, alias_dict, overriding_kargs)

##__________________________________________________________________||
def _expand_str(path_cfg, alias_dict, overriding_kargs):
    """expand a path config given as a string

    """

    if path_cfg in alias_dict:
        # e.g., path_cfg = 'var_cut'
        return _expand_str_alias(path_cfg, alias_dict, overriding_kargs)

    # e.g., path_cfg = 'ev : {low} <= ev.var[0] < {high}'
    return _expand_for_lambda_str(path_cfg, alias_dict, overriding_kargs)

def _expand_for_lambda_str(path_cfg, alias_dict, overriding_kargs):

    # e.g.,
    # path_cfg = 'ev : {low} <= ev.var[0] < {high}'

    ret = dict(factory='LambdaStrFactory', lambda_str=path_cfg, components=())
    # e.g.,
    # {
    #     'factory': 'LambdaStrFactory',
    #     'lambda_str': 'ev : {low} <= ev.var[0] < {high}'
    # }

    overriding_kargs_copy = overriding_kargs.copy()
    # e.g., {'low': 25, 'high': 200, 'alias': 'var_cut', 'name': 'var_cut25'}

    if 'alias' in overriding_kargs:
        ret['name'] = overriding_kargs_copy.pop('alias')

    if 'name' in overriding_kargs:
        ret['name'] = overriding_kargs_copy.pop('name')

    ret.update(overriding_kargs_copy)
    # e.g.,
    # {
    #     'factory': 'LambdaStrFactory',
    #     'lambda_str': 'ev : {low} <= ev.var[0] < {high}',
    #     'name': 'var_cut25',
    #     'low': 25, 'high': 200
    # }

    return ret

def _expand_str_alias(path_cfg, alias_dict, overriding_kargs):
    """expand a path config given as a string

    Args:
        path_cfg (str): an alias
        alias_dict (dict):
        overriding_kargs (dict):
    """

    # e.g.,
    # path_cfg = 'var_cut'

    new_path_cfg = alias_dict[path_cfg]
    # e.g., ('ev : {low} <= ev.var[0] < {high}', {'low': 10, 'high': 200})

    new_overriding_kargs = dict(alias=path_cfg)
    # e.g., {'alias': 'var_cut'}

    new_overriding_kargs.update(overriding_kargs)
    # e.g., {'alias': 'var_cut', 'name': 'var_cut25', 'low': 25}

    return expand_path_cfg(new_path_cfg, alias_dict,new_overriding_kargs)

##__________________________________________________________________||
def _expand_tuple(path_cfg, alias_dict, overriding_kargs):
    """expand a path config given as a tuple

    """

    # e.g.,
    # path_cfg = ('ev : {low} <= ev.var[0] < {high}', {'low': 10, 'high': 200})
    # overriding_kargs = {'alias': 'var_cut', 'name': 'var_cut25', 'low': 25}

    new_path_cfg = path_cfg[0]
    # e.g., 'ev : {low} <= ev.var[0] < {high}'

    new_overriding_kargs = path_cfg[1].copy()
    # e.g., {'low': 10, 'high': 200}

    new_overriding_kargs.update(overriding_kargs)
    # e.g., {'low': 25, 'high': 200, 'alias': 'var_cut', 'name': 'var_cut25'}

    return expand_path_cfg(
        new_path_cfg,
        overriding_kargs=new_overriding_kargs,
        alias_dict=alias_dict
    )


##__________________________________________________________________||
def _expand_dict(path_cfg, alias_dict):
    if 'factory' in path_cfg:
        return path_cfg

    if not sum([k in path_cfg for k in ('All', 'Any', 'Not')]) <= 1:
        raise ValueError("Any pair of 'All', 'Any', 'Not' cannot be simultaneously given unless factory is given!")

    if 'All' in path_cfg:
        new_path_cfg = path_cfg.copy()
        new_path_cfg['factory'] = 'AllFactory'
        new_path_cfg['components'] = tuple([expand_path_cfg(p, alias_dict=alias_dict) for p in new_path_cfg.pop('All')])
        return new_path_cfg

    if 'Any' in path_cfg:
        new_path_cfg = path_cfg.copy()
        new_path_cfg['factory'] = 'AnyFactory'
        new_path_cfg['components'] = tuple([expand_path_cfg(p, alias_dict=alias_dict) for p in new_path_cfg.pop('Any')])
        return new_path_cfg

    if 'Not' in path_cfg:
        new_path_cfg = path_cfg.copy()
        new_path_cfg['factory'] = 'NotFactory'
        new_path_cfg['components'] = (expand_path_cfg(new_path_cfg.pop('Not'), alias_dict=alias_dict), )
        return new_path_cfg

    raise ValueError("cannot recognize the path_cfg")

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
class DatasetReaderComposite(object):

    def __init__(self):
        self.readers = [ ]

    def __repr__(self):
        name_value_pairs = (
            ('readers', self.readers),
        )
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(n, v) for n, v in name_value_pairs]),
        )

    def add(self, reader):
        self.readers.append(reader)

    def begin(self):
        for reader in self.readers:
            reader.begin()

    def read(self, dataset):
        for reader in self.readers:
            reader.read(dataset)

    def end(self):
        return [reader.end() for reader in self.readers]

##__________________________________________________________________||



import unittest

import copy
import numpy as np

from alphatwirl.summary import Sum

##__________________________________________________________________||
class TestSum(unittest.TestCase):

    def test_init(self):
        obj = Sum()
        np.testing.assert_equal([np.array([0])], obj.contents)

    def test_init_val(self):
        obj = Sum(val = (10, 20))
        np.testing.assert_equal([np.array([10, 20])], obj.contents)

    def test_init_weight(self):
        obj = Sum(val = (10, 20), weight = 2)
        np.testing.assert_equal([np.array([20, 40])], obj.contents)

    def test_init_contents(self):
        obj = Sum(contents = [np.array((1, 3))])
        np.testing.assert_equal([np.array([1, 3])], obj.contents)

    @unittest.skip("they are the same object now")
    def test_init_contents_not_same_object(self):
        contents = [np.array([1, 3])]
        obj = Sum(contents = contents)
        np.testing.assert_equal(contents, obj.contents)
        self.assertIsNot(contents, obj.contents)
        self.assertIsNot(contents[0], obj.contents[0])

    def test_repr(self):
        obj = Sum()
        repr(obj)

    def test_add(self):
        obj1 = Sum(contents = [np.array((10, 20))])
        obj2 = Sum(contents = [np.array((30, 40))])
        obj3 =  obj1 + obj2
        np.testing.assert_equal([np.array([40, 60])], obj3.contents)
        self.assertIsNot(obj1, obj3)
        self.assertIsNot(obj1.contents, obj3.contents)
        self.assertIsNot(obj2, obj3)
        self.assertIsNot(obj2.contents, obj3.contents)

    def test_add_int_to_float_val(self):
        obj1 = Sum(contents = [np.array((10, 20))])
        obj2 = Sum(contents = [np.array((30, 40.2))])
        obj3 =  obj1 + obj2
        np.testing.assert_equal([np.array([40, 60.2])], obj3.contents)
        self.assertIsNot(obj1, obj3)
        self.assertIsNot(obj1.contents, obj3.contents)
        self.assertIsNot(obj2, obj3)
        self.assertIsNot(obj2.contents, obj3.contents)

    def test_radd(self):
        obj1 = Sum(contents = [np.array((10, 20))])
        self.assertIsNot(obj1, sum([obj1])) # will call 0 + obj1
        self.assertEqual(obj1, sum([obj1]))

    def test_radd_raise(self):
        obj1 = Sum(contents = [np.array((10, 20))])
        self.assertRaises(TypeError, obj1.__radd__, 1 ) # 1 + obj1

    def test_copy(self):
        obj1 = Sum(contents = [np.array((10, 20))])
        copy1 = copy.copy(obj1)
        self.assertEqual(obj1, copy1)
        self.assertIsNot(obj1, copy1)
        self.assertIsNot(obj1.contents, copy1.contents)
        self.assertIsNot(obj1.contents[0], copy1.contents[0])

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

import re

##__________________________________________________________________||
def quote_string(text):

    if not text: return '""'

    to_quote = False

    if re.search(r'"', text):
        text = re.sub(r'"', r'\"', text) # escape double quote with backslash
        to_quote = True

    if ' ' in text:
        to_quote = True

    if to_quote:
        text = '"{}"'.format(text)

    return text

##__________________________________________________________________||
